"""This module provides classes to specify atomic structure."""

import numpy as np
from jarvis.core.composition import Composition
from jarvis.core.specie import Specie, atomic_numbers_to_symbols
from jarvis.core.lattice import Lattice, lattice_coords_transformer
from collections import OrderedDict
from jarvis.core.utils import get_counts
import itertools
from jarvis.core.utils import (
    check_duplicate_coords,
    get_new_coord_for_xyz_sym,
    gcd,
    get_angle,
)
import os
import math
import tempfile
import random
import string
import datetime
from collections import defaultdict

amu_gm = 1.66054e-24
ang_cm = 1e-8


class Atoms(object):
    """
    Generate Atoms python object.

    >>> box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    >>> coords = [[0, 0, 0], [0.25, 0.2, 0.25]]
    >>> elements = ["Si", "Si"]
    >>> Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    >>> print(round(Si.volume,2))
    40.03
    >>> Si.composition
    {'Si': 2}
    >>> round(Si.density,2)
    2.33
    >>> round(Si.packing_fraction,2)
    0.28
    >>> Si.atomic_numbers
    [14, 14]
    >>> Si.num_atoms
    2
    >>> Si.frac_coords[0][0]
    0
    >>> Si.cart_coords[0][0]
    0.0
    >>> coords = [[0, 0, 0], [1.3575 , 1.22175, 1.22175]]
    >>> round(Si.density,2)
    2.33
    >>> Si.spacegroup()
    'C2/m (12)'
    >>> Si.pymatgen_converter()!={}
    True
    """

    def __init__(
        self,
        lattice_mat=None,
        coords=None,
        elements=None,
        props=None,
        cartesian=False,
        show_props=False,
    ):
        """
        Create atomic structure.

        Requires lattice, coordinates, atom type  information.
        """
        self.lattice_mat = np.array(lattice_mat)
        self.show_props = show_props
        self.lattice = Lattice(lattice_mat)
        self.coords = np.array(coords)
        self.elements = elements
        self.cartesian = cartesian
        self.props = props
        if self.props is None:
            self.props = ["" for i in range(len(self.elements))]
        if self.cartesian:
            self.cart_coords = self.coords
            self.frac_coords = np.array(self.lattice.frac_coords(self.coords))
        else:
            self.frac_coords = self.coords
            self.cart_coords = np.array(self.lattice.cart_coords(self.coords))

    def write_cif(
        self, filename="atoms.cif", comment=None, with_spg_info=True
    ):
        """
        Write CIF format file from Atoms object.

        Caution: can't handle fractional occupancies right now
        """
        if comment is None:
            comment = "CIF file written using JARVIS-Tools package."
        comment = comment + "\n"
        f = open(filename, "w")
        f.write(comment)
        composition = self.composition
        line = "data_" + str(composition.reduced_formula) + "\n"
        f.write(line)
        from jarvis.analysis.structure.spacegroup import Spacegroup3D

        if with_spg_info:
            spg = Spacegroup3D(self)
            line = (
                "_symmetry_space_group_name_H-M  "
                + str("'")
                + str(spg.space_group_symbol)
                + str("'")
                + "\n"
            )
        else:
            line = "_symmetry_space_group_name_H-M  " + str("'P 1'") + "\n"

        f.write(line)
        a, b, c, alpha, beta, gamma = self.lattice.parameters
        f.write("_cell_length_a       %g\n" % a)
        f.write("_cell_length_b       %g\n" % b)
        f.write("_cell_length_c       %g\n" % c)
        f.write("_cell_angle_alpha    %g\n" % alpha)
        f.write("_cell_angle_beta     %g\n" % beta)
        f.write("_cell_angle_gamma    %g\n" % gamma)
        f.write("\n")
        if with_spg_info:
            line = (
                "_symmetry_Int_Tables_number  "
                + str(spg.space_group_number)
                + "\n"
            )
        else:
            line = "_symmetry_Int_Tables_number  " + str(1) + "\n"
        f.write(line)

        line = (
            "_chemical_formula_structural  "
            + str(composition.reduced_formula)
            + "\n"
        )
        f.write(line)
        line = "_chemical_formula_sum  " + str(composition.formula) + "\n"
        f.write(line)
        line = "_cell_volume  " + str(self.volume) + "\n"
        f.write(line)
        reduced, repeat = composition.reduce()
        line = "_cell_formula_units_Z  " + str(repeat) + "\n"
        f.write(line)
        f.write("loop_\n")
        f.write("  _symmetry_equiv_pos_site_id\n")
        f.write(" _symmetry_equiv_pos_as_xyz\n")
        f.write(" 1  'x, y, z'\n")
        f.write("loop_\n")
        f.write(" _atom_site_type_symbol\n")
        f.write(" _atom_site_label\n")
        f.write(" _atom_site_symmetry_multiplicity\n")
        f.write(" _atom_site_fract_x\n")
        f.write(" _atom_site_fract_y\n")
        f.write(" _atom_site_fract_z\n")
        f.write(" _atom_site_fract_occupancy\n")
        order = np.argsort(self.elements)
        coords_ordered = np.array(self.frac_coords)[order]
        elements_ordered = np.array(self.elements)[order]
        occ = 1
        extra = 1
        element_types = []
        # count = 0
        for ii, i in enumerate(elements_ordered):
            if i not in element_types:
                element_types.append(i)
                count = 0
            symbol = i
            count = count + 1
            label = str(i) + str(count)
            element_types.append(i)
            coords = coords_ordered[ii]
            f.write(
                " %s  %s  %s  %7.5f  %7.5f  %7.5f  %s\n"
                % (symbol, label, occ, coords[0], coords[1], coords[2], extra)
            )
        f.close()

    @staticmethod
    def read_with_cif2cell(filename="1000000.cif", get_primitive_atoms=False):
        """Use cif2cell package to read cif files."""
        # https://pypi.org/project/cif2cell/
        # tested on version 2.0.0a3
        try:
            new_file, fname = tempfile.mkstemp(text=True)
            cmd = (
                "cif2cell "
                + filename
                + " -p vasp --vasp-cartesian-positions --vca -o "
                + fname
            )
            os.system(cmd)

        except Exception as exp:
            print(exp)
            pass
        f = open(fname, "r")
        text = f.read().splitlines()
        f.close()
        os.close(new_file)
        elements = text[0].split("Species order:")[1].split()
        scale = float(text[1])
        lattice_mat = []
        lattice_mat.append([float(i) for i in text[2].split()])
        lattice_mat.append([float(i) for i in text[3].split()])
        lattice_mat.append([float(i) for i in text[4].split()])
        lattice_mat = scale * np.array(lattice_mat)
        uniq_elements = elements
        element_count = np.array([int(i) for i in text[5].split()])
        elements = []
        for i, ii in enumerate(element_count):
            for j in range(ii):
                elements.append(uniq_elements[i])
        cartesian = True
        if "d" in text[6] or "D" in text[6]:
            cartesian = False
        # print ('cartesian poscar=',cartesian,text[7])
        num_atoms = int(np.sum(element_count))
        coords = []
        for i in range(num_atoms):
            coords.append([float(i) for i in text[7 + i].split()[0:3]])
        coords = np.array(coords)
        atoms = Atoms(
            lattice_mat=lattice_mat,
            coords=coords,
            elements=elements,
            cartesian=cartesian,
        )
        if get_primitive_atoms:
            atoms = atoms.get_primitive_atoms
        return atoms

    @staticmethod
    def from_pdb_old(filename="abc.pdb"):
        """Read PDB file, kept of checking, use from_pdb instead."""
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        coords = []
        species = []
        for i in lines:
            tmp = i.split()
            if "ATOM " in i and "REMARK" not in i and "SITE" not in i:
                coord = [float(tmp[6]), float(tmp[7]), float(tmp[8])]
                coords.append(coord)
                species.append(tmp[11])
                # print (coord,tmp[11])
        coords = np.array(coords)

        max_c = np.max(coords, axis=0)
        min_c = np.min(coords, axis=0)
        box = np.zeros((3, 3))
        for j in range(3):
            box[j, j] = abs(max_c[j] - min_c[j])
        pdb = Atoms(
            lattice_mat=box, elements=species, coords=coords, cartesian=True
        )
        mol = VacuumPadding(pdb, vacuum=20.0).get_effective_molecule()
        return mol

    @staticmethod
    def from_pdb(filename="abc.pdb", max_lat=200):
        """Read pdb/sdf/mol2 etc. file and make Atoms object, using pytraj."""
        try:
            import pytraj as pt
        except Exception:
            print("Pytraj not installed, trying native version.")
            return Atoms.from_pdb_old(filename)
        x = pt.load(filename)
        coords = x.xyz
        at_numbs = [int(i.atomic_number) for i in list(x.top.atoms)]
        elements = atomic_numbers_to_symbols(at_numbs)
        lattice_mat = [[max_lat, 0, 0], [0, max_lat, 0], [0, 0, max_lat]]
        a = Atoms(
            lattice_mat=lattice_mat,
            elements=elements,
            coords=coords[0],
            cartesian=True,
        )
        return a

    @staticmethod
    def from_cif(
        filename="atoms.cif",
        from_string="",
        get_primitive_atoms=True,
        use_cif2cell=True,
    ):
        """Read .cif format file."""
        # Warnings:
        # May not work for:
        # system with partial occupancy
        # cif file with multiple blocks
        # _atom_site_U_iso, instead of fractn_x, cartn_x
        # with non-zero _atom_site_attached_hydrogens
        try:
            if use_cif2cell:
                # https://pypi.org/project/cif2cell/
                # tested on version 2.0.0a3
                if from_string != "":
                    new_file, filename = tempfile.mkstemp(text=True)
                    f = open(filename, "w")
                    f.write(from_string)
                    f.close()
                atoms = Atoms.read_with_cif2cell(
                    filename=filename, get_primitive_atoms=get_primitive_atoms
                )

                return atoms
        except Exception as exp:
            print(exp)
            pass

        if from_string == "":
            f = open(filename, "r")
            lines = f.read().splitlines()
            # lines = [ii.encode('utf-8') for ii in f.read().splitlines()]
            f.close()
        else:
            lines = from_string.splitlines()
        lat_a = ""
        lat_b = ""
        lat_c = ""
        lat_alpha = ""
        lat_beta = ""
        lat_gamma = ""
        # TODO: check if chemical_formula_sum
        # matches Atoms.compsotion.reduced_formula

        # chemical_formula_structural = ""
        # chemical_formula_sum = ""
        # chemical_name_mineral = ""
        sym_xyz_line = ""
        for ii, i in enumerate(lines):
            if "_cell_length_a" in i:
                lat_a = float(i.split()[1].split("(")[0])
            if "_cell_length_b" in i:
                lat_b = float(i.split()[1].split("(")[0])
            if "_cell_length_c" in i:
                lat_c = float(i.split()[1].split("(")[0])
            if "_cell_angle_alpha" in i:
                lat_alpha = float(i.split()[1].split("(")[0])
            if "_cell_angle_beta" in i:
                lat_beta = float(i.split()[1].split("(")[0])
            if "_cell_angle_gamma" in i:
                lat_gamma = float(i.split()[1].split("(")[0])
            # if "_chemical_formula_structural" in i:
            #    chemical_formula_structural = i.split()[1]
            # if "_chemical_formula_sum" in i:
            #    chemical_formula_sum = i.split()[1]
            # if "_chemical_name_mineral" in i:
            #    chemical_name_mineral = i.split()[1]
            if "_symmetry_equiv_pos_as_xyz" in i:
                sym_xyz_line = ii
            if "_symmetry_equiv_pos_as_xyz_" in i:
                sym_xyz_line = ii
            if "_symmetry_equiv_pos_as_xyz_" in i:
                sym_xyz_line = ii
            if "_space_group_symop_operation_xyz_" in i:
                sym_xyz_line = ii
            if "_space_group_symop_operation_xyz" in i:
                sym_xyz_line = ii
        symm_ops = []
        terminate = False
        count = 0
        while not terminate:
            # print("sym_xyz_line", sym_xyz_line)
            tmp = lines[sym_xyz_line + count + 1]
            if "x" in tmp and "y" in tmp and "z" in tmp:
                # print("tmp", tmp)
                symm_ops.append(tmp)
                count += 1
            else:
                terminate = True
        tmp_arr = [lat_a, lat_b, lat_c, lat_alpha, lat_beta, lat_gamma]
        if any(ele == "" for ele in tmp_arr):
            raise ValueError("Lattice information is incomplete.", tmp_arr)
        lat = Lattice.from_parameters(
            lat_a, lat_b, lat_c, lat_alpha, lat_beta, lat_gamma
        )
        terminate = False
        atom_features = []
        count = 0
        beginning_atom_info_line = 0
        for ii, i in enumerate(lines):
            if "loop_" in i and "_atom_site" in lines[ii + count + 1]:
                beginning_atom_info_line = ii
        while not terminate:
            if "_atom" in lines[beginning_atom_info_line + count + 1]:
                atom_features.append(
                    lines[beginning_atom_info_line + count + 1]
                )
            count += 1
            if "_atom" not in lines[beginning_atom_info_line + count]:
                terminate = True
        terminate = False
        count = 1
        atom_liines = []
        while not terminate:
            number = beginning_atom_info_line + len(atom_features) + count
            if number == len(lines):
                terminate = True
                break
            line = lines[number]
            # print ('tis line',line)
            if len(line.split()) == len(atom_features):
                atom_liines.append(line)
                count += 1
            else:
                terminate = True
        label_index = ""
        fract_x_index = ""
        fract_y_index = ""
        fract_z_index = ""
        cartn_x_index = ""
        cartn_y_index = ""
        cartn_z_index = ""
        occupancy_index = ""
        for ii, i in enumerate(atom_features):
            if "_atom_site_label" in i:
                label_index = ii
            if "fract_x" in i:
                fract_x_index = ii
            if "fract_y" in i:
                fract_y_index = ii
            if "fract_z" in i:
                fract_z_index = ii
            if "cartn_x" in i:
                cartn_x_index = ii
            if "cartn_y" in i:
                cartn_y_index = ii
            if "cartn_z" in i:
                cartn_z_index = ii
            if "occupancy" in i:
                occupancy_index = ii
        if fract_x_index == "" and cartn_x_index == "":
            raise ValueError("Cannot find atomic coordinate info.")
        elements = []
        coords = []
        cif_atoms = None
        if fract_x_index != "":
            for ii, i in enumerate(atom_liines):
                tmp = i.split()
                tmp_lbl = list(
                    Composition.from_string(tmp[label_index]).to_dict().keys()
                )
                elem = tmp_lbl[0]
                coord = [
                    float(tmp[fract_x_index].split("(")[0]),
                    float(tmp[fract_y_index].split("(")[0]),
                    float(tmp[fract_z_index].split("(")[0]),
                ]
                if len(tmp_lbl) > 1:
                    raise ValueError("Check if labesl are correct.", tmp_lbl)
                if (
                    occupancy_index != ""
                    and not float(
                        tmp[occupancy_index].split("(")[0]
                    ).is_integer()
                ):
                    raise ValueError(
                        "Fractional occupancy is not supported.",
                        float(tmp[occupancy_index].split("(")[0]),
                        elem,
                    )

                elements.append(elem)
                coords.append(coord)
            cif_atoms = Atoms(
                lattice_mat=lat.matrix,
                elements=elements,
                coords=coords,
                cartesian=False,
            )
        elif cartn_x_index != "":
            for ii, i in enumerate(atom_liines):
                tmp = i.split()
                tmp_lbl = list(
                    Composition.from_string(tmp[label_index]).to_dict().keys()
                )
                elem = tmp_lbl[0]
                coord = [
                    float(tmp[cartn_x_index].split("(")[0]),
                    float(tmp[cartn_y_index].split("(")[0]),
                    float(tmp[cartn_z_index].split("(")[0]),
                ]
                if len(tmp_lbl) > 1:
                    raise ValueError("Check if labesl are correct.", tmp_lbl)
                if (
                    occupancy_index != ""
                    and not float(
                        tmp[occupancy_index].split("(")[0]
                    ).is_integer()
                ):
                    raise ValueError(
                        "Fractional occupancy is not supported.",
                        float(tmp[occupancy_index].split("(")[0]),
                        elem,
                    )
                elements.append(elem)
                coords.append(coord)
            cif_atoms = Atoms(
                lattice_mat=lat.matrix,
                elements=elements,
                coords=coords,
                cartesian=True,
            )
        else:
            raise ValueError(
                "Cannot find atomic coordinate info from cart or frac."
            )
        # frac_coords=list(cif_atoms.frac_coords)
        cif_elements = cif_atoms.elements
        lat = cif_atoms.lattice.matrix
        if len(symm_ops) > 1:
            frac_coords = list(cif_atoms.frac_coords)
            for i in symm_ops:
                for jj, j in enumerate(frac_coords):
                    new_c_coord = get_new_coord_for_xyz_sym(
                        xyz_string=i, frac_coord=j
                    )
                    new_frac_coord = [new_c_coord][0]
                    if not check_duplicate_coords(frac_coords, new_frac_coord):
                        frac_coords.append(new_frac_coord)
                        cif_elements.append(cif_elements[jj])
            new_atoms = Atoms(
                lattice_mat=lat,
                coords=frac_coords,
                elements=cif_elements,
                cartesian=False,
            )
            cif_atoms = new_atoms
        if get_primitive_atoms:
            cif_atoms = cif_atoms.get_primitive_atoms
        return cif_atoms

    def write_poscar(self, filename="POSCAR"):
        """Write POSCAR format file from Atoms object."""
        from jarvis.io.vasp.inputs import Poscar

        pos = Poscar(self)
        pos.write_file(filename)

    @property
    def get_xyz_string(self):
        """Get xyz string for atoms."""
        line = str(self.num_atoms) + "\n"
        line += " ".join(map(str, np.array(self.lattice_mat).flatten())) + "\n"
        for i, j in zip(self.elements, self.cart_coords):
            line += (
                str(i)
                + str(" ")
                + str(round(j[0], 4))
                + str(" ")
                + str(round(j[1], 4))
                + str(" ")
                + str(round(j[2], 4))
                + "\n"
            )
        return line

    def write_xyz(self, filename="atoms.xyz"):
        """Write XYZ format file."""
        f = open(filename, "w")
        line = str(self.num_atoms) + "\n"
        f.write(line)
        line = ",".join(map(str, np.array(self.lattice_mat).flatten())) + "\n"
        f.write(line)
        for i, j in zip(self.elements, self.cart_coords):
            f.write("%s %7.5f %7.5f %7.5f\n" % (i, j[0], j[1], j[2]))
        f.close()

    @classmethod
    def from_xyz(self, filename="dsgdb9nsd_057387.xyz", box_size=40):
        """Read XYZ file from to make Atoms object."""
        lattice_mat = [[box_size, 0, 0], [0, box_size, 0], [0, 0, box_size]]
        f = open(filename, "r")
        lines = f.read().splitlines()
        f.close()
        try:
            lattice_mat = np.array(lines[1].split(","), dtype="float").reshape(
                3, 3
            )
        except Exception:
            pass
        coords = []
        species = []
        natoms = int(lines[0])
        for i in range(natoms):
            tmp = (lines[i + 2]).split()
            coord = [(tmp[1]), (tmp[2]), (tmp[3])]
            coord = [
                0 if "*" in ii else float(ii) for ii in coord
            ]  # dsgdb9nsd_000212.xyz
            coords.append(coord)
            species.append(tmp[0])
        coords = np.array(coords)
        atoms = Atoms(
            lattice_mat=lattice_mat,
            coords=coords,
            elements=species,
            cartesian=True,
        ).center_around_origin(new_origin=[0.5, 0.5, 0.5])
        # print (atoms)
        return atoms

    @classmethod
    def from_poscar(self, filename="POSCAR"):
        """Read POSCAR/CONTCAR file from to make Atoms object."""
        from jarvis.io.vasp.inputs import Poscar

        return Poscar.from_file(filename).atoms

    @property
    def check_polar(self):
        """
        Check if the surface structure is polar.

        Comparing atom types at top and bottom.
        Applicable for sufcae with vaccums only.

        Args:

             file:atoms object (surface with vacuum)

        Returns:
               polar:True/False
        """
        up = 0
        dn = 0
        tol = 0.01
        coords = np.array(self.frac_coords) % 1
        z_max = max(coords[:, 2])
        z_min = min(coords[:, 2])
        for site, element in zip(coords, self.elements):
            if site[2] >= z_max - tol:
                # if site[2] == z_max:
                up = up + Specie(element).Z
            if site[2] <= z_min + tol:
                # if site[2] == z_min:
                dn = dn + Specie(element).Z
        polar = False
        if up != dn:
            print("Seems like a polar materials.")
            polar = True
        if up == dn:
            print("Non-polar")
            polar = False
        return polar

    def strain_atoms(self, strain):
        """Apply volumetric strain to atoms."""
        # strains=np.arange(0.80,1.2,0.02)
        s = np.eye(3) + np.array(strain) * np.eye(3)
        lattice_mat = np.array(self.lattice_mat)
        lattice_mat = np.dot(lattice_mat, s)
        return Atoms(
            lattice_mat=lattice_mat,
            elements=self.elements,
            coords=self.coords,
            cartesian=self.cartesian,
        )

    def apply_strain(self, strain):
        """Apply a strain(e.g. 0.01, [0,0,.01]) to the lattice."""
        print("Use strain_atoms instead.")
        s = (1 + np.array(strain)) * np.eye(3)
        self.lattice_mat = np.dot(self.lattice_mat.T, s).T

    def to_dict(self):
        """Provide dictionary representation of the atoms object."""
        d = OrderedDict()
        d["lattice_mat"] = self.lattice_mat.tolist()
        d["coords"] = np.array(self.coords).tolist()
        d["elements"] = np.array(self.elements).tolist()
        d["abc"] = self.lattice.lat_lengths()
        d["angles"] = self.lattice.lat_angles()
        d["cartesian"] = self.cartesian
        d["props"] = self.props
        return d

    @classmethod
    def from_dict(self, d={}):
        """Form atoms object from the dictionary."""
        return Atoms(
            lattice_mat=d["lattice_mat"],
            elements=d["elements"],
            props=d["props"],
            coords=d["coords"],
            cartesian=d["cartesian"],
        )

    def remove_site_by_index(self, site=0):
        """Remove an atom by its index number."""
        return self.remove_sites_by_indices(indices=[site])

    def remove_sites_by_indices(self, indices=[0], in_place=False):
        """Remove multiple atoms by their corresponding indices number."""
        new_els = []
        new_coords = []
        new_props = []
        for ii, i in enumerate(self.frac_coords):
            if ii not in indices:
                new_els.append(self.elements[ii])
                new_coords.append(self.frac_coords[ii])
                new_props.append(self.props[ii])
        if in_place:
            self.elements = new_els
            self.coords = new_coords
            self.props = new_props
            return self
        return Atoms(
            lattice_mat=self.lattice_mat,
            elements=new_els,
            coords=np.array(new_coords),
            props=new_props,
            cartesian=False,
        )

    def add_site(
        self, element="Si", coords=[0.1, 0.1, 0.1], props=[], index=0
    ):
        """Ad an atom, coords in fractional coordinates."""
        new_els = list(self.elements)
        new_coords = list(self.frac_coords)
        new_props = list(self.props)
        new_els.insert(index, element)
        new_coords.insert(index, coords)
        new_props.insert(index, props)
        # new_els.append(element)
        # new_coords.append(coords)
        # new_props.append(props)
        new_els = np.array(new_els)
        new_coords = np.array(new_coords)
        new_props = np.array(new_props, dtype=object)
        return Atoms(
            lattice_mat=self.lattice_mat,
            elements=new_els,
            coords=np.array(new_coords),
            props=new_props,
            cartesian=False,
        )

    @property
    def get_conventional_atoms(self):
        """Get conventional Atoms using spacegroup information."""
        from jarvis.analysis.structure.spacegroup import Spacegroup3D

        return Spacegroup3D(self).conventional_standard_structure

    @property
    def get_spacegroup(self):
        """Get spacegroup information."""
        from jarvis.analysis.structure.spacegroup import Spacegroup3D

        spg = Spacegroup3D(self)
        return [spg.space_group_number, spg.space_group_symbol]

    @property
    def get_primitive_atoms(self):
        """Get primitive Atoms using spacegroup information."""
        from jarvis.analysis.structure.spacegroup import Spacegroup3D

        return Spacegroup3D(self).primitive_atoms

    def get_all_neighbors(self, r=5, bond_tol=0.15):
        """
        Get neighbors for each atom in the unit cell, out to a distance r.

        Contains [index_i, index_j, distance, image] array.
        Adapted from pymatgen.
        """
        recp_len = np.array(self.lattice.reciprocal_lattice().abc)
        maxr = np.ceil((r + bond_tol) * recp_len / (2 * math.pi))
        nmin = np.floor(np.min(self.frac_coords, axis=0)) - maxr
        nmax = np.ceil(np.max(self.frac_coords, axis=0)) + maxr
        all_ranges = [np.arange(x, y) for x, y in zip(nmin, nmax)]
        matrix = self.lattice_mat
        neighbors = [list() for _ in range(len(self.cart_coords))]
        # all_fcoords = np.mod(self.frac_coords, 1)
        coords_in_cell = self.cart_coords  # np.dot(all_fcoords, matrix)
        site_coords = self.cart_coords
        indices = np.arange(len(site_coords))
        for image in itertools.product(*all_ranges):
            coords = np.dot(image, matrix) + coords_in_cell
            z = (coords[:, None, :] - site_coords[None, :, :]) ** 2
            all_dists = np.sum(z, axis=-1) ** 0.5
            all_within_r = np.bitwise_and(all_dists <= r, all_dists > 1e-8)
            for j, d, within_r in zip(indices, all_dists, all_within_r):
                for i in indices[within_r]:
                    if d[i] > bond_tol:
                        # if d[i] > bond_tol and i!=j:
                        neighbors[i].append([i, j, d[i], image])
        return np.array(neighbors, dtype="object")

    def get_neighbors_cutoffs(self, max_cut=10, r=5, bond_tol=0.15):
        """Get neighbors within cutoff."""
        neighbors = self.get_all_neighbors(r=r, bond_tol=bond_tol)
        dists = np.hstack(([[xx[2] for xx in yy] for yy in neighbors]))
        hist, bins = np.histogram(dists, bins=np.arange(0.1, 10.2, 0.1))
        # shell_vol = (
        #    4.0
        #    / 3.0
        #    * np.pi
        #    * (np.power(bins[1:], 3) - np.power(bins[:-1], 3))
        # )
        arr = []
        for i, j in zip(bins[:-1], hist):
            if j > 0:
                arr.append(i)
        rcut_buffer = 0.11
        io1 = 0
        io2 = 1
        io3 = 2
        try:
            delta = arr[io2] - arr[io1]
            while delta < rcut_buffer and arr[io2] < max_cut:
                io1 = io1 + 1
                io2 = io2 + 1
                io3 = io3 + 1
                delta = arr[io2] - arr[io1]

            rcut1 = (arr[io2] + arr[io1]) / float(2.0)
        except Exception:
            print("Warning:Setting first nbr cut-off as minimum bond-dist")
            rcut1 = arr[0]
            pass
        try:
            delta = arr[io3] - arr[io2]
            while (
                delta < rcut_buffer
                and arr[io3] < max_cut
                and arr[io2] < max_cut
            ):
                io2 = io2 + 1
                io3 = io3 + 1
                delta = arr[io3] - arr[io2]
            rcut2 = float(arr[io3] + arr[io2]) / float(2.0)
        except Exception:
            print("Warning:Setting first and second nbr cut-off equal")
            print("You might consider increasing max_n parameter")
            rcut2 = rcut1
            pass
        return rcut1, rcut2, neighbors

    def atomwise_angle_and_radial_distribution(
        self, r=5, bond_tol=0.15, c_size=10, verbose=False
    ):
        """Get atomwise distributions."""
        rcut1, rcut2, neighbors = self.get_neighbors_cutoffs(
            r=r, bond_tol=bond_tol
        )
        from jarvis.analysis.structure.neighbors import NeighborsAnalysis
        from jarvis.core.utils import bond_angle as angle

        nbor_info = NeighborsAnalysis(
            self, rcut1=rcut1, rcut2=rcut2
        ).nbor_list(rcut=rcut1, c_size=c_size)
        nat = nbor_info["nat"]
        dist = nbor_info["dist"]
        atom_rdfs = []
        nbins = 180
        actual_prdf = []
        # actual_pangs = []
        actual_pangs = np.zeros((nat, 380))
        for i in range(nat):
            hist, bins = np.histogram(
                dist[:, i], bins=np.arange(0.1, rcut1 + 0.2, 0.1)
            )
            actual_prdf.append(dist[:, i])
            atom_rdfs.append(hist.tolist())
            if verbose:
                exact_dists = np.arange(0.1, rcut1 + 0.2, 0.1)[hist.nonzero()]
                print("exact_dists", exact_dists)
        # prdf_arr = np.array(atom_rdfs)[0 : self.num_atoms]

        atom_angles = []
        for i in range(nbor_info["nat"]):
            angles = [
                angle(
                    nbor_info["dist"][in1][i],
                    nbor_info["dist"][in2][i],
                    nbor_info["bondx"][in1][i],
                    nbor_info["bondx"][in2][i],
                    nbor_info["bondy"][in1][i],
                    nbor_info["bondy"][in2][i],
                    nbor_info["bondz"][in1][i],
                    nbor_info["bondz"][in2][i],
                )
                for in1 in range(nbor_info["nn"][i])
                for in2 in range(nbor_info["nn"][i])
                if in2 > in1
                and nbor_info["dist"][in1][i] * nbor_info["dist"][in2][i] != 0
            ]
            ang_hist, ang_bins = np.histogram(
                angles,
                bins=np.arange(1, nbins + 2, 1),
                density=False,
            )
            for jj, j in enumerate(angles):
                actual_pangs[i, jj] = j
            # actual_pangs.append(angles)
            atom_angles.append(ang_hist)
            if verbose:
                exact_angles = np.arange(1, nbins + 2, 1)[ang_hist.nonzero()]
                print("exact_angles", exact_angles)
        # return (atom_angles)#/nbor_info['nat']

        # pangle_arr = np.array(atom_angles)[0 : self.num_atoms]
        return (
            neighbors,
            np.array(atom_rdfs)[0 : self.num_atoms],
            np.array(atom_angles)[0 : self.num_atoms],
            np.array(actual_prdf[0 : self.num_atoms]),
            np.array(actual_pangs[0 : self.num_atoms]),
            nbor_info,
        )

    @property
    def raw_distance_matrix(self):
        """Provide distance matrix."""
        coords = np.array(self.cart_coords)
        z = (coords[:, None, :] - coords[None, :, :]) ** 2
        return np.sum(z, axis=-1) ** 0.5

    @property
    def raw_angle_matrix(self, cut_off=5.0):
        """Provide distance matrix."""
        coords = np.array(self.cart_coords)
        angles = []
        for a, b, c in itertools.product(coords, coords, coords):
            if (
                not np.array_equal(a, b)
                and not np.array_equal(b, c)
                and np.linalg.norm((a - b)) < cut_off
                and np.linalg.norm((c - b)) < cut_off
            ):
                angle = get_angle(a, b, c)
                angles.append(angle)
        return angles

    def center(self, axis=2, vacuum=18.0, about=None):
        """
        Center structure with vacuum padding.

        Args:
          vacuum:vacuum size

          axis: direction
        """
        cell = self.lattice_mat
        p = self.cart_coords
        props = self.props
        dirs = np.zeros_like(cell)
        for i in range(3):
            dirs[i] = np.cross(cell[i - 1], cell[i - 2])
            dirs[i] /= np.sqrt(np.dot(dirs[i], dirs[i]))  # normalize
            if np.dot(dirs[i], cell[i]) < 0.0:
                dirs[i] *= -1

        if isinstance(axis, int):
            axes = (axis,)
        else:
            axes = axis

        # if vacuum and any(self.pbc[x] for x in axes):
        #     warnings.warn(
        #         'You are adding vacuum along a periodic direction!')

        # Now, decide how much each basis vector should be made longer
        longer = np.zeros(3)
        shift = np.zeros(3)
        for i in axes:
            p0 = np.dot(p, dirs[i]).min() if len(p) else 0
            p1 = np.dot(p, dirs[i]).max() if len(p) else 0
            height = np.dot(cell[i], dirs[i])
            if vacuum is not None:
                lng = (p1 - p0 + 2 * vacuum) - height
            else:
                lng = 0.0  # Do not change unit cell size!
            top = lng + height - p1
            shf = 0.5 * (top - p0)
            cosphi = np.dot(cell[i], dirs[i]) / np.sqrt(
                np.dot(cell[i], cell[i])
            )
            longer[i] = lng / cosphi
            shift[i] = shf / cosphi

        # Now, do it!
        translation = np.zeros(3)
        for i in axes:
            nowlen = np.sqrt(np.dot(cell[i], cell[i]))
            if vacuum is not None or cell[i].any():
                cell[i] = cell[i] * (1 + longer[i] / nowlen)
                translation += shift[i] * cell[i] / nowlen

        new_coords = p + translation
        if about is not None:
            for vector in cell:
                new_coords -= vector / 2.0
            new_coords += about
        atoms = Atoms(
            lattice_mat=cell,
            elements=self.elements,
            coords=new_coords,
            cartesian=True,
            props=props,
        )
        return atoms

    @property
    def volume(self):
        """Get volume of the atoms object."""
        m = self.lattice_mat
        vol = float(abs(np.dot(np.cross(m[0], m[1]), m[2])))
        return vol

    @property
    def composition(self):
        """Get composition of the atoms object."""
        comp = {}
        for i in self.elements:
            comp[i] = comp.setdefault(i, 0) + 1
        return Composition(OrderedDict(comp))

    @property
    def density(self):
        """Get density in g/cm3 of the atoms object."""
        den = float(self.composition.weight * amu_gm) / (
            float(self.volume) * (ang_cm) ** 3
        )
        return den

    @property
    def atomic_numbers(self):
        """Get list of atomic numbers of atoms in the atoms object."""
        numbers = []
        for i in self.elements:
            numbers.append(Specie(i).Z)
        return numbers

    @property
    def num_atoms(self):
        """Get number of atoms."""
        if np.squeeze(self.coords).ndim == 1:
            return 1
        return len(self.coords)

    @property
    def uniq_species(self):
        """Get unique elements."""
        uniq = set()
        return [x for x in self.elements if not (x in uniq or uniq.add(x))]

    def get_center_of_mass(self):
        """Get center of mass of the atoms object."""
        # atomic_mass
        m = []
        for i in self.elements:
            m.append(Specie(i).atomic_mass)
        m = np.array(m)
        com = np.dot(m, self.cart_coords) / m.sum()
        # com = np.linalg.solve(self.lattice_mat.T, com)
        return com

    def get_origin(self):
        """Get center of mass of the atoms object."""
        # atomic_mass
        return self.frac_coords.mean(axis=0)

    def center_around_origin(self, new_origin=[0.0, 0.0, 0.5]):
        """Center around given origin."""
        lat = self.lattice_mat
        typ_sp = self.elements
        natoms = self.num_atoms
        props = self.props
        # abc = self.lattice.lat_lengths()
        COM = self.get_origin()
        # COM = self.get_center_of_mass()
        x = np.zeros((natoms))
        y = np.zeros((natoms))
        z = np.zeros((natoms))
        coords = list()
        for i in range(0, natoms):
            # cart_coords[i]=self.cart_coords[i]-COM
            x[i] = self.frac_coords[i][0] - COM[0] + new_origin[0]
            y[i] = self.frac_coords[i][1] - COM[1] + new_origin[1]
            z[i] = self.frac_coords[i][2] - COM[2] + new_origin[2]
            coords.append([x[i], y[i], z[i]])
        struct = Atoms(
            lattice_mat=lat,
            elements=typ_sp,
            coords=coords,
            cartesian=False,
            props=props,
        )
        return struct

    def pymatgen_converter(self):
        """Get pymatgen representation of the atoms object."""
        try:
            from pymatgen.core.structure import Structure

            return Structure(
                self.lattice_mat,
                self.elements,
                self.frac_coords,
                coords_are_cartesian=False,
            )
        except Exception:
            print("Requires pymatgen for this functionality.")
            pass

    def phonopy_converter(self, pbc=True):
        """Get phonopy representation of the atoms object."""
        try:
            from phonopy.structure.atoms import Atoms as PhonopyAtoms

            return PhonopyAtoms(
                symbols=self.elements,
                positions=self.cart_coords,
                pbc=pbc,
                cell=self.lattice_mat,
            )
        except Exception:
            print("Requires phonopy for this functionality.")
            pass

    def ase_converter(self, pbc=True):
        """Get ASE representation of the atoms object."""
        try:
            from ase import Atoms as AseAtoms

            return AseAtoms(
                symbols=self.elements,
                positions=self.cart_coords,
                pbc=pbc,
                cell=self.lattice_mat,
            )
        except Exception:
            print("Requires ASE for this functionality.")
            pass

    def spacegroup(self, symprec=1e-3):
        """Get spacegroup of the atoms object."""
        import spglib

        sg = spglib.get_spacegroup(
            (self.lattice_mat, self.frac_coords, self.atomic_numbers),
            symprec=symprec,
        )
        return sg

    @property
    def packing_fraction(self):
        """Get packing fraction of the atoms object."""
        total_rad = 0
        for i in self.elements:
            total_rad = total_rad + Specie(i).atomic_rad ** 3
        pf = np.array([4 * np.pi * total_rad / (3 * self.volume)])
        return round(pf[0], 5)

    def lattice_points_in_supercell(self, supercell_matrix):
        """
        Adapted from Pymatgen.

        Returns the list of points on the original lattice contained in the
        supercell in fractional coordinates (with the supercell basis).
        e.g. [[2,0,0],[0,1,0],[0,0,1]] returns [[0,0,0],[0.5,0,0]]

        Args:

            supercell_matrix: 3x3 matrix describing the supercell

        Returns:
            numpy array of the fractional coordinates
        """
        diagonals = np.array(
            [
                [0, 0, 0],
                [0, 0, 1],
                [0, 1, 0],
                [0, 1, 1],
                [1, 0, 0],
                [1, 0, 1],
                [1, 1, 0],
                [1, 1, 1],
            ]
        )
        d_points = np.dot(diagonals, supercell_matrix)

        mins = np.min(d_points, axis=0)
        maxes = np.max(d_points, axis=0) + 1

        ar = (
            np.arange(mins[0], maxes[0])[:, None]
            * np.array([1, 0, 0])[None, :]
        )
        br = (
            np.arange(mins[1], maxes[1])[:, None]
            * np.array([0, 1, 0])[None, :]
        )
        cr = (
            np.arange(mins[2], maxes[2])[:, None]
            * np.array([0, 0, 1])[None, :]
        )

        all_points = ar[:, None, None] + br[None, :, None] + cr[None, None, :]
        all_points = all_points.reshape((-1, 3))

        frac_points = np.dot(all_points, np.linalg.inv(supercell_matrix))

        tvects = frac_points[
            np.all(frac_points < 1 - 1e-10, axis=1)
            & np.all(frac_points >= -1e-10, axis=1)
        ]
        assert len(tvects) == round(abs(np.linalg.det(supercell_matrix)))
        return tvects

    def describe(
        self,
        xrd_peaks=5,
        xrd_round=1,
        cutoff=4,
        take_n_bonds=2,
        include_spg=True,
    ):
        """Describe for NLP applications."""
        from jarvis.analysis.diffraction.xrd import XRD

        if include_spg:
            from jarvis.analysis.structure.spacegroup import Spacegroup3D

            spg = Spacegroup3D(self)
        theta, d_hkls, intens = XRD().simulate(atoms=self)
        #     x = atoms.atomwise_angle_and_radial_distribution()
        #     bond_distances = {}
        #     for i, j in x[-1]["different_bond"].items():
        #         bond_distances[i.replace("_", "-")] = ", ".join(
        #             map(str, (sorted(list(set([round(jj, 2) for jj in j])))))
        #         )
        dists = defaultdict(list)
        elements = self.elements
        for i in self.get_all_neighbors(r=cutoff):
            for j in i:
                key = "-".join(sorted([elements[j[0]], elements[j[1]]]))
                dists[key].append(j[2])
        bond_distances = {}
        for i, j in dists.items():
            dist = sorted(set([round(k, 2) for k in j]))
            if len(dist) >= take_n_bonds:
                dist = dist[0:take_n_bonds]
            bond_distances[i] = ", ".join(map(str, dist))
        fracs = {}
        for i, j in (self.composition.atomic_fraction).items():
            fracs[i] = round(j, 3)
        info = {}
        chem_info = {
            "atomic_formula": self.composition.reduced_formula,
            "prototype": self.composition.prototype,
            "molecular_weight": round(self.composition.weight / 2, 2),
            "atomic_fraction": (fracs),
            "atomic_X": ", ".join(
                map(str, [Specie(s).X for s in self.uniq_species])
            ),
            "atomic_Z": ", ".join(
                map(str, [Specie(s).Z for s in self.uniq_species])
            ),
        }
        struct_info = {
            "lattice_parameters": ", ".join(
                map(str, [round(j, 2) for j in self.lattice.abc])
            ),
            "lattice_angles": ", ".join(
                map(str, [round(j, 2) for j in self.lattice.angles])
            ),
            # "spg_number": spg.space_group_number,
            # "spg_symbol": spg.space_group_symbol,
            "top_k_xrd_peaks": ", ".join(
                map(
                    str,
                    sorted(list(set([round(i, xrd_round) for i in theta])))[
                        0:xrd_peaks
                    ],
                )
            ),
            "density": round(self.density, 3),
            # "crystal_system": spg.crystal_system,
            # "point_group": spg.point_group_symbol,
            # "wyckoff": ", ".join(list(set(spg._dataset["wyckoffs"]))),
            "bond_distances": bond_distances,
        }
        if include_spg:
            struct_info["spg_number"] = spg.space_group_number
            struct_info["spg_symbol"] = spg.space_group_symbol
            struct_info["crystal_system"] = spg.crystal_system
            struct_info["point_group"] = spg.point_group_symbol
            struct_info["wyckoff"] = ", ".join(
                list(set(spg._dataset["wyckoffs"]))
            )
            struct_info["natoms_primitive"] = spg.primitive_atoms.num_atoms
            struct_info[
                "natoms_conventional"
            ] = spg.conventional_standard_structure.num_atoms
        info["chemical_info"] = chem_info
        info["structure_info"] = struct_info
        line = "The number of atoms are: " + str(
            self.num_atoms
        )  # +"., The elements are: "+",".join(atoms.elements)+". "
        for i, j in info.items():
            if not isinstance(j, dict):
                line += "The " + i + " is " + j + ". "
            else:
                # print("i",i)
                # print("j",j)
                for ii, jj in j.items():
                    tmp = ""
                    if isinstance(jj, dict):
                        for iii, jjj in jj.items():
                            tmp += iii + ": " + str(jjj) + " "
                    else:
                        tmp = jj
                    line += "The " + ii + " is " + str(tmp) + ". "
        line1 = line
        # print('bond_distances', struct_info['bond_distances'])
        tmp = ""
        p = struct_info["bond_distances"]
        for ii, (kk, vv) in enumerate(p.items()):
            if ii == len(p) - 1:
                punc = " Å."
            else:
                punc = " Å; "
            tmp += kk + ": " + vv + punc
        line2 = (
            chem_info["atomic_formula"]
            + " crystallizes in the "
            + struct_info["crystal_system"]
            + " "
            + str(struct_info["spg_symbol"])
            + " spacegroup, "
            + struct_info["point_group"]
            + " pointgroup with a prototype of "
            + str(chem_info["prototype"])
            # " and a molecular weight of " +
            # str(chem_info['molecular_weight']) +
            + ". The atomic fractions are: "
            + str(chem_info["atomic_fraction"])
            .replace("{", "")
            .replace("}", "")
            + " with electronegaticities as "
            + str(chem_info["atomic_X"])
            + " and atomic numbers as "
            + str(chem_info["atomic_Z"])
            + ". The bond distances are: "
            + str(tmp)
            + "The lattice lengths are: "
            + struct_info["lattice_parameters"]
            + " Å, and the lattice angles are: "
            + struct_info["lattice_angles"]
            + "º with some of the top XRD peaks at "
            + struct_info["top_k_xrd_peaks"]
            + "º with "
            + "Wyckoff symbols "
            + struct_info["wyckoff"]
            + "."
        )
        info["desc_1"] = line1
        info["desc_2"] = line2
        return info

    def make_supercell_matrix(self, scaling_matrix):
        """
        Adapted from Pymatgen.

        Makes a supercell. Allowing to have sites outside the unit cell.

        Args:
            scaling_matrix: A scaling matrix for transforming the lattice
            vectors. Has to be all integers. Several options are possible:
            a. A full 3x3 scaling matrix defining the linear combination
             the old lattice vectors. E.g., [[2,1,0],[0,3,0],[0,0,
             1]] generates a new structure with lattice vectors a' =
             2a + b, b' = 3b, c' = c where a, b, and c are the lattice
             vectors of the original structure.
            b. An sequence of three scaling factors. E.g., [2, 1, 1]
             specifies that the supercell should have dimensions 2a x b x
             c.
            c. A number, which simply scales all lattice vectors by the
             same factor.

        Returns:
            Supercell structure. Note that a Structure is always returned,
            even if the input structure is a subclass of Structure. This is
            to avoid different arguments signatures from causing problems. If
            you prefer a subclass to return its own type, you need to override
            this method in the subclass.
        """
        scale_matrix = np.array(scaling_matrix, np.int16)
        if scale_matrix.shape != (3, 3):
            scale_matrix = np.array(scale_matrix * np.eye(3), np.int16)
        new_lattice = Lattice(np.dot(scale_matrix, self.lattice_mat))

        f_lat = self.lattice_points_in_supercell(scale_matrix)
        c_lat = new_lattice.cart_coords(f_lat)

        new_sites = []
        new_elements = []
        for site, el in zip(self.cart_coords, self.elements):
            for v in c_lat:
                new_elements.append(el)
                tmp = site + v
                new_sites.append(tmp)
        return Atoms(
            lattice_mat=new_lattice.lattice(),
            elements=new_elements,
            coords=new_sites,
            cartesian=True,
        )

    def make_supercell(self, dim=[2, 2, 2]):
        """Make supercell of dimension dim."""
        dim = np.array(dim)
        if dim.shape == (3, 3):
            dim = np.array([int(np.linalg.norm(v)) for v in dim])
        return self.make_supercell_matrix(dim)

    def make_supercell_old(self, dim=[2, 2, 2]):
        """Make supercell of dimension dim using for loop."""
        coords = self.frac_coords
        all_symbs = self.elements  # [i.symbol for i in s.species]
        nat = len(coords)

        new_nat = nat * dim[0] * dim[1] * dim[2]
        new_coords = np.zeros((new_nat, 3))
        new_symbs = []  # np.chararray((new_nat))
        props = []  # self.props

        ct = 0
        for i in range(nat):
            for j in range(dim[0]):
                for k in range(dim[1]):
                    for m in range(dim[2]):
                        props.append(self.props[i])
                        new_coords[ct][0] = (coords[i][0] + j) / float(dim[0])
                        new_coords[ct][1] = (coords[i][1] + k) / float(dim[1])
                        new_coords[ct][2] = (coords[i][2] + m) / float(dim[2])
                        new_symbs.append(all_symbs[i])
                        ct = ct + 1

        nat = new_nat

        nat = len(coords)  # int(s.composition.num_atoms)
        lat = np.zeros((3, 3))
        box = self.lattice_mat
        lat[0][0] = dim[0] * box[0][0]
        lat[0][1] = dim[0] * box[0][1]
        lat[0][2] = dim[0] * box[0][2]
        lat[1][0] = dim[1] * box[1][0]
        lat[1][1] = dim[1] * box[1][1]
        lat[1][2] = dim[1] * box[1][2]
        lat[2][0] = dim[2] * box[2][0]
        lat[2][1] = dim[2] * box[2][1]
        lat[2][2] = dim[2] * box[2][2]
        super_cell = Atoms(
            lattice_mat=lat,
            coords=new_coords,
            elements=new_symbs,
            props=props,
            cartesian=False,
        )
        return super_cell

    def get_lll_reduced_structure(self):
        """Get LLL algorithm based reduced structure."""
        reduced_latt = self.lattice.get_lll_reduced_lattice()
        if reduced_latt != self.lattice:
            return Atoms(
                lattice_mat=reduced_latt._lat,
                elements=self.elements,
                coords=self.frac_coords,
                cartesian=False,
            )
        else:
            return Atoms(
                lattice_mat=self.lattice_mat,
                elements=self.elements,
                coords=self.frac_coords,
                cartesian=False,
            )

    def __repr__(self):
        """Get representation during print statement."""
        from jarvis.io.vasp.inputs import Poscar

        return Poscar(self).to_string()
        # return self.get_string()

    def get_string(self, cart=True, sort_order="X"):
        """
        Convert Atoms to string.

        Optional arguments below.

        Args:
          cart:True/False for cartesian/fractional coords.

          sort_order: sort by chemical properties of
                    elements. Default electronegativity.
        """
        from jarvis.io.vasp.inputs import Poscar

        return Poscar(self).to_string()
        # Might be a bug while sorting

        system = str(self.composition.formula)
        header = (
            str(system)
            + str("\n1.0\n")
            + str(self.lattice_mat[0][0])
            + " "
            + str(self.lattice_mat[0][1])
            + " "
            + str(self.lattice_mat[0][2])
            + "\n"
            + str(self.lattice_mat[1][0])
            + " "
            + str(self.lattice_mat[1][1])
            + " "
            + str(self.lattice_mat[1][2])
            + "\n"
            + str(self.lattice_mat[2][0])
            + " "
            + str(self.lattice_mat[2][1])
            + " "
            + str(self.lattice_mat[2][2])
            + "\n"
        )
        if sort_order is None:
            order = np.argsort(self.elements)
        else:
            order = np.argsort(
                [Specie(i).element_property(sort_order) for i in self.elements]
            )
        if cart:
            coords = self.cart_coords
        else:
            coords = self.frac_coords
        coords_ordered = np.array(coords)[order]
        elements_ordered = np.array(self.elements)[order]
        props_ordered = np.array(self.props)[order]
        # check_selective_dynamics = False # TODO
        counts = get_counts(elements_ordered)
        cart_frac = ""
        if cart:
            cart_frac = "\nCartesian\n"
        else:
            cart_frac = "\nDirect\n"

        if "T" in "".join(map(str, self.props[0])):
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + "\nSelective dynamics\n"
                + cart_frac
            )
        else:
            middle = (
                " ".join(map(str, counts.keys()))
                + "\n"
                + " ".join(map(str, counts.values()))
                + cart_frac
            )
        rest = ""
        if coords_ordered.ndim == 1:
            coords_ordered = np.array([coords])
        for ii, i in enumerate(coords_ordered):
            if self.show_props:
                rest = (
                    rest
                    + " ".join(map(str, i))
                    + " "
                    + str(props_ordered[ii])
                    + "\n"
                )
            else:
                rest = rest + " ".join(map(str, i)) + "\n"

        result = header + middle + rest
        return result

    def clone(self):
        """Clones the class instance."""
        return Atoms(
            lattice_mat=self.lattice_mat,
            elements=self.elements,
            coords=self.frac_coords,
            props=self.props,
            cartesian=self.cartesian,
            show_props=self.show_props,
        )


class VacuumPadding(object):
    """Adds vaccum padding to make 2D structure or making molecules."""

    def __init__(self, atoms, vacuum=20.0):
        """
        Initialize object.

        Args:
            atoms: Atoms object
            vacuum:  vacuum padding
        """
        self.atoms = atoms
        self.vacuum = vacuum

    def get_effective_2d_slab(self):
        """Add 2D vacuum to a system."""
        z_coord = []
        for i in self.atoms.frac_coords:
            tmp = i[2]
            if i[2] >= 0.5:
                tmp = i[2] - 1
            elif i[2] < -0.5:
                tmp = i[2] + 1
            z_coord.append(tmp)

        zmaxp = max(np.array(z_coord) * self.atoms.lattice_mat[2][2])
        zminp = min(np.array(z_coord) * self.atoms.lattice_mat[2][2])
        thickness = abs(zmaxp - zminp)
        padding = self.vacuum + thickness

        # lattice_mat = self.atoms.lattice_mat
        # lattice_mat[2][2] = padding
        # elements = self.atoms.elements
        # atoms = atoms.center(vacuum=0.0)
        new_lat = self.atoms.lattice_mat
        a1 = new_lat[0]
        a2 = new_lat[1]
        a3 = new_lat[2]
        new_lat = np.array(
            [
                a1,
                a2,
                np.cross(a1, a2)
                * np.dot(a3, np.cross(a1, a2))
                / np.linalg.norm(np.cross(a1, a2)) ** 2,
            ]
        )

        a1 = new_lat[0]
        a2 = new_lat[1]
        a3 = new_lat[2]

        latest_lat = np.array(
            [
                (np.linalg.norm(a1), 0, 0),
                (
                    np.dot(a1, a2) / np.linalg.norm(a1),
                    np.sqrt(
                        np.linalg.norm(a2) ** 2
                        - (np.dot(a1, a2) / np.linalg.norm(a1)) ** 2
                    ),
                    0,
                ),
                (0, 0, np.linalg.norm(a3)),
            ]
        )
        # latest_lat[2][2] = padding
        M = np.linalg.solve(new_lat, latest_lat)
        new_cart_coords = self.atoms.cart_coords
        new_coords = np.dot(new_cart_coords, M)
        new_atoms = Atoms(
            lattice_mat=latest_lat,
            elements=self.atoms.elements,
            coords=new_coords,
            cartesian=True,
        )
        frac_coords = new_atoms.frac_coords
        frac_coords[:] = frac_coords[:] % 1
        new_atoms = Atoms(
            lattice_mat=latest_lat,
            elements=self.atoms.elements,
            coords=frac_coords,
            cartesian=False,
        )
        new_lat = new_atoms.lattice_mat
        new_cart_coords = new_atoms.cart_coords
        elements = new_atoms.elements
        new_lat[2][2] = padding  # new_lat[2][2]+ self.vacuum
        with_vacuum_atoms = Atoms(
            lattice_mat=new_lat,
            elements=elements,
            coords=new_cart_coords,
            cartesian=True,
        )
        frac = np.array(with_vacuum_atoms.frac_coords)
        frac[:, 2] = frac[:, 2] - np.mean(frac[:, 2]) + 0.5
        frac[:, 2] = frac[:, 2] - np.mean(frac[:, 2]) + 0.5
        with_vacuum_atoms = Atoms(
            lattice_mat=new_lat,
            elements=elements,
            coords=frac,
            cartesian=False,
        )
        return with_vacuum_atoms

    def get_effective_molecule(self):
        """Add vacuum around a system."""
        x_coord = []
        y_coord = []
        z_coord = []
        for i in self.atoms.frac_coords:
            tmp = i[0]
            if i[0] >= 0.5:
                tmp = i[0] - 1
            elif i[0] < -0.5:
                tmp = i[0] + 1
            x_coord.append(tmp)
            tmp = i[1]
            if i[1] >= 0.5:
                tmp = i[1] - 1
            elif i[1] < -0.5:
                tmp = i[1] + 1
            y_coord.append(tmp)
            tmp = i[2]
            if i[2] >= 0.5:
                tmp = i[2] - 1
            elif i[2] < -0.5:
                tmp = i[2] + 1
            z_coord.append(tmp)

        xmaxp = max(np.array(x_coord) * self.atoms.lattice_mat[0][0])
        xminp = min(np.array(x_coord) * self.atoms.lattice_mat[0][0])
        ymaxp = max(np.array(y_coord) * self.atoms.lattice_mat[1][1])
        yminp = min(np.array(y_coord) * self.atoms.lattice_mat[1][1])
        zmaxp = max(np.array(z_coord) * self.atoms.lattice_mat[2][2])
        zminp = min(np.array(z_coord) * self.atoms.lattice_mat[2][2])
        thickness_x = abs(xmaxp - xminp)
        thickness_y = abs(ymaxp - yminp)
        thickness_z = abs(zmaxp - zminp)

        lattice_mat = np.zeros((3, 3))  # self.atoms.lattice_mat
        lattice_mat[0][0] = self.vacuum + thickness_x
        lattice_mat[1][1] = self.vacuum + thickness_y
        lattice_mat[2][2] = self.vacuum + thickness_z
        elements = self.atoms.elements
        atoms = Atoms(
            lattice_mat=lattice_mat,
            coords=self.atoms.cart_coords,
            elements=elements,
            cartesian=True,
        )
        frac = np.array(atoms.frac_coords)
        frac[:, 0] = frac[:, 0] - np.mean(frac[:, 0]) + 0.5
        frac[:, 1] = frac[:, 1] - np.mean(frac[:, 1]) + 0.5
        frac[:, 2] = frac[:, 2] - np.mean(frac[:, 2]) + 0.5
        with_vacuum_atoms = Atoms(
            lattice_mat=lattice_mat,
            elements=elements,
            coords=frac,
            cartesian=False,
        )
        return with_vacuum_atoms


def fix_pbc(atoms):
    """Use for making Atoms with vacuum."""
    new_f_coords = []
    for i in atoms.frac_coords:
        if i[2] > 0.5:
            i[2] = i[2] - 1
        if i[2] < -0.5:
            i[2] = i[2] + 1
        new_f_coords.append(i)
    return Atoms(
        lattice_mat=atoms.lattice_mat,
        elements=atoms.elements,
        coords=new_f_coords,
        cartesian=False,
    )


def add_atoms(
    top, bottom, distance=[0, 0, 1], apply_strain=False, add_tags=True
):
    """
    Add top and bottom Atoms with a distance array.

    Bottom Atoms lattice-matrix is chosen as final lattice.
    """
    top = top.center_around_origin([0, 0, 0])
    bottom = bottom.center_around_origin(distance)
    strain_x = (
        top.lattice_mat[0][0] - bottom.lattice_mat[0][0]
    ) / bottom.lattice_mat[0][0]
    strain_y = (
        top.lattice_mat[1][1] - bottom.lattice_mat[1][1]
    ) / bottom.lattice_mat[1][1]
    if apply_strain:
        top.apply_strain([strain_x, strain_y, 0])
    #  print("strain_x,strain_y", strain_x, strain_y)
    elements = []
    coords = []
    props = []
    lattice_mat = bottom.lattice_mat
    for i, j in zip(bottom.elements, bottom.frac_coords):
        elements.append(i)
        coords.append(j)
        props.append("bottom")
    top_cart_coords = lattice_coords_transformer(
        new_lattice_mat=top.lattice_mat,
        old_lattice_mat=bottom.lattice_mat,
        cart_coords=top.cart_coords,
    )
    top_frac_coords = bottom.lattice.frac_coords(top_cart_coords)
    for i, j in zip(top.elements, top_frac_coords):
        elements.append(i)
        coords.append(j)
        props.append("top")

    order = np.argsort(np.array(elements))
    elements = np.array(elements)[order]
    coords = np.array(coords)[order]
    determnt = np.linalg.det(np.array(lattice_mat))
    if determnt < 0.0:
        lattice_mat = -1 * np.array(lattice_mat)
    determnt = np.linalg.det(np.array(lattice_mat))
    if determnt < 0.0:
        print("Serious issue, check lattice vectors.")
        print("Many software follow right hand basis rule only.")
    combined = Atoms(
        lattice_mat=lattice_mat,
        coords=coords,
        elements=elements,
        props=props,
        cartesian=False,
    ).center_around_origin()
    # print('combined props',combined)
    return combined


def get_supercell_dims(atoms, enforce_c_size=10, extend=1):
    """Get supercell dimensions."""
    a = atoms.lattice.lat_lengths()[0]
    b = atoms.lattice.lat_lengths()[1]
    c = atoms.lattice.lat_lengths()[2]
    dim1 = int(float(enforce_c_size) / float(a)) + extend
    dim2 = int(float(enforce_c_size) / float(b)) + extend
    dim3 = int(float(enforce_c_size) / float(c)) + extend
    return [dim1, dim2, dim3]


def pmg_to_atoms(pmg=""):
    """Convert pymatgen structure to Atoms."""
    return Atoms(
        lattice_mat=pmg.lattice.matrix,
        elements=[i.symbol for i in pmg.species],
        coords=pmg.frac_coords,
        cartesian=False,
    )


def ase_to_atoms(ase_atoms="", cartesian=True):
    """Convert ase structure to Atoms."""
    return Atoms(
        lattice_mat=ase_atoms.get_cell(),
        elements=ase_atoms.get_chemical_symbols(),
        coords=ase_atoms.get_positions(),
        cartesian=cartesian,
    )


def crop_square(atoms=None, csize=10):
    """Crop a sqaur portion from a surface/2D system."""
    sz = csize / 2
    # just to make sure we have enough material to crop from
    enforce_c_size = sz * 3
    dims = get_supercell_dims(atoms, enforce_c_size=enforce_c_size)
    b = atoms.make_supercell_matrix(dims).center_around_origin()
    lat_mat = [
        [enforce_c_size, 0, 0],
        [0, enforce_c_size, 0],
        [0, 0, b.lattice_mat[2][2]],
    ]
    # M = np.linalg.solve(b.lattice_mat, lat_mat)
    # tol = 3

    els = []
    coords = []
    for i, j in zip(b.cart_coords, b.elements):
        if i[0] <= sz and i[0] >= -sz and i[1] <= sz and i[1] >= -sz:
            els.append(j)
            coords.append(i)
    coords = np.array(coords)
    # new_mat = (
    #    [max(coords[:, 0]) - min(coords[:, 0]) + tol, 0, 0],
    #    [0, max(coords[:, 1]) - min(coords[:, 1]) + tol, 0],
    #    [0, 0, b.lattice_mat[2][2]],
    # )
    new_atoms = Atoms(
        lattice_mat=lat_mat, elements=els, coords=coords, cartesian=True
    ).center_around_origin([0.5, 0.5, 0.5])
    return new_atoms


class OptimadeAdaptor(object):
    """Module to work with optimade."""

    def __init__(self, atoms=None):
        """Intialize class with Atoms object."""
        self.atoms = atoms

    def reduce(self, content={}):
        """Reduce chemical formula."""
        repeat = 0
        for specie, count in content.items():
            if repeat == 0:
                repeat = count
            else:
                repeat = gcd(count, repeat)
        reduced = {}
        for specie, count in content.items():
            reduced[specie] = count // repeat
        return reduced, repeat

    def optimade_reduced_formula(self, content={}, sort_alphabetical=True):
        """Get chemical formula."""
        if sort_alphabetical:
            content = OrderedDict(
                sorted(content.items(), key=lambda x: (x[0]))
            )

        form = ""
        reduced, repeat = self.reduce(content)
        Z = {}
        for i, j in reduced.items():
            Z[i] = reduced[i]
        for specie, count in Z.items():
            if float(count).is_integer():
                if count == 1:
                    form = form + specie
                else:
                    form = form + specie + str(int(count))
            else:
                form = form + specie + str(count)

        return form  # .replace("1", "")

    def get_optimade_prototype(self, content={}):
        """Get chemical prototypes such as A, AB etc."""
        reduced, repeat = self.reduce(content)
        proto = ""
        all_upper = string.ascii_uppercase
        items = sorted(list(reduced.values()), reverse=True)
        N = 0
        # for specie, count in reduced.items():
        for c in items:
            proto = proto + str(all_upper[N]) + str(round(c, 3))
            N = N + 1
        return proto.replace("1", "")

    def get_optimade_species(self):
        """Get optimade species."""
        atoms = self.atoms
        elements = np.array(list(set(atoms.elements)))
        order = np.argsort(elements)
        elements = (elements)[order]
        sp = []
        for i in elements:
            info = {}
            info["name"] = i
            info["chemical_symbols"] = [i]
            info["concentration"] = [1.0]
            info["mass"] = None
            info["original_name"] = None
            info["attached"] = None
            info["nattached"] = None
            sp.append(info)
        return sp

    def from_optimade(self, info={}):
        """Get Atoms from optimade."""
        lattice_mat = info["attributes"]["lattice_vectors"]
        elements = info["attributes"]["elements"]
        coords = info["attributes"]["cartesian_site_positions"]
        return Atoms(
            lattice_mat=lattice_mat,
            elements=elements,
            coords=coords,
            cartesian=True,
        )

    def to_optimade(
        self,
        idx="x",
        itype="structures",
        source="JARVIS-DFT-3D",
        reference_url="https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/",
        now=datetime.datetime.now(datetime.timezone.utc).isoformat(),
    ):
        """Get optimade format data."""
        atoms = self.atoms
        info = {}
        info["id"] = idx
        info["type"] = itype
        info_at = {}
        info_at["_jarvis_source"] = source
        info_at["_jarvis_reference"] = (
            reference_url
            + idx
            # + ".xml"
        )
        comp = atoms.composition.to_dict()
        info_at["chemical_formula_reduced"] = self.optimade_reduced_formula(
            comp
        )
        info_at["chemical_formula_anonymous"] = self.get_optimade_prototype(
            comp
        )
        elements = np.array(atoms.elements)
        order = np.argsort(elements)
        info_at["elements"] = elements[order]
        info_at["nelements"] = len(elements)
        info_at["nsites"] = len(atoms.frac_coords)
        info_at["lattice_vectors"] = atoms.lattice_mat.tolist()
        info_at["species_at_sites"] = np.array(atoms.elements)[order]
        info_at["cartesian_site_positions"] = atoms.cart_coords[order].tolist()
        info_at["nperiodic_dimensions"] = 3
        # info_at["species"] = atoms.elements
        info_at[
            "species"
        ] = self.get_optimade_species()  # dict(atoms.composition.to_dict())
        info_at["elements_ratios"] = list(
            atoms.composition.atomic_fraction.values()
        )
        info_at["structure_features"] = []
        info_at["last_modified"] = str(now)
        # info_at["more_data_available"] = True
        info_at[
            "chemical_formula_descriptive"
        ] = atoms.composition.reduced_formula
        info_at["dimension_types"] = [1, 1, 1]
        info["attributes"] = info_at
        return info


def compare_atoms(atoms1=[], atoms2=[], primitive_cell=True, verbose=True):
    """Compare atomic strutures."""
    from jarvis.analysis.structure.spacegroup import Spacegroup3D

    if primitive_cell:
        atoms1 = atoms1.get_primitive_atoms
        atoms2 = atoms2.get_primitive_atoms
    formula1 = atoms1.composition.reduced_formula
    formula2 = atoms2.composition.reduced_formula
    if formula1 != formula2:
        if verbose:
            print("Formula dont match", formula1, formula2)
        return False
    if atoms1.num_atoms != atoms2.num_atoms:
        if verbose:
            print("Num_atoms dont match", atoms1.num_atoms, atoms2.num_atoms)
        return False
    spg1 = Spacegroup3D(atoms1).space_group_number
    spg2 = Spacegroup3D(atoms2).space_group_number
    if spg1 != spg2:
        if verbose:
            print("Spg dont match", spg1, spg2)
        return False
    return True


def build_xanes_poscar(
    atoms=None,
    selected_element="Si",
    prefix="-",
    extend=1,
    enforce_c_size=12,
    dir=".",
    filename_with_prefix=False,
):
    """Generate POSCAR file for XANES, note the element ordering."""
    # from jarvis.core.utils import rand_select
    from jarvis.analysis.structure.spacegroup import Spacegroup3D

    dims = get_supercell_dims(
        atoms, enforce_c_size=enforce_c_size, extend=extend
    )
    # spg = Spacegroup3D(atoms)
    # wyckoffs = spg._dataset["wyckoffs"]
    # atoms.props = wyckoffs
    atoms = atoms.make_supercell_matrix(dims)
    spath = os.path.join(dir, "POSCAR-supercell.vasp")
    atoms.write_poscar(spath)
    spg = Spacegroup3D(atoms)
    wyckoffs = spg._dataset["wyckoffs"]
    atoms.props = wyckoffs
    # print ('atoms.props',atoms.props)
    el_props = []
    elements = atoms.elements
    for ii, i in enumerate(elements):
        if i == selected_element:
            el_props.append(ii)
    choice = random.choice(el_props)
    props = {atoms.props[choice]: choice}
    tmp_atoms = atoms
    for i, j in props.items():
        if tmp_atoms.elements[j] == selected_element:
            atoms = tmp_atoms
            defect_strt = atoms.remove_site_by_index(j)
            coords = tmp_atoms.frac_coords[j]
            added_strt = defect_strt.add_site(element="XANES", coords=coords)
            if filename_with_prefix:
                filename = (
                    "POSCAR"
                    + prefix
                    + tmp_atoms.elements[j]
                    + "_"
                    + str(j)
                    + "_"
                    + tmp_atoms.props[j]
                    + ".vasp"
                )
            else:
                filename = "POSCAR"

            filename = os.path.join(dir, filename)
            added_strt.props = ["" for i in range(added_strt.num_atoms)]
            added_strt.write_poscar(filename)
            with open(filename, "r") as f:
                filedata = f.read()
            newdata = filedata.replace("XANES", tmp_atoms.elements[j])
            with open(filename, "w") as f:
                f.write(newdata)

    return atoms


# ['Mn ', 'Mn ', 'Ru ', 'U ']
#
# def clear_elements(atoms=None):
# return info
#    info={}
#    info['lattice_mat']=atoms['lattice_mat']
#    info['coords']=atoms['coords']
#    info['props']=atoms['props']
#    info['cartesian']=atoms['cartesian']
#    elements=[i.strip() for i in atoms['elements']]
#    info['elements']=elements
#    return info



"""Run multiple jobs."""
from jarvis.tasks.qe.super import SuperCond
from jarvis.core.utils import get_factors
from jarvis.core.atoms import Atoms
from jarvis.db.figshare import data
from jarvis.core.kpoints import Kpoints3D
from jarvis.tasks.queue_jobs import Queue
from jarvis.db.jsonutils import dumpjson
import os
from jarvis.analysis.structure.spacegroup import Spacegroup3D

# import glob
# from jarvis.db.jsonutils import loadjson

my_data = data("dft_3d")


def get_jid_data(jid="JVASP-667", dataset="dft_2d"):
    """Get info for a jid and dataset."""
    # d = data(dataset)
    d = my_data
    for i in d:
        if i["jid"] == jid:
            return i


qe_cmd = "mpirun -np 16 /home/kfg/codes/q-e-qe-6.5/bin/pw.x"
qe_cmd = "/home/kfg/codes/q-e-qe-6.5/bin/pw.x"

run_dir = "/wrk/knc6/Super"
run_dir = "/wrk/knc6/CDVAE_SUP"


def non_prime_kpoints(kpts=[]):
    """Get non prime kpoints."""
    mem = []
    for i in kpts:
        facts = get_factors(i)
        if len(facts) == 1:
            val = i + 1
        else:
            val = i
        mem.append(val)
    return mem


def write_qejob(pyname="job.py", job_json=""):
    """Write template job.py with VaspJob.to_dict() job.json."""
    f = open(pyname, "w")
    f.write("from jarvis.tasks.qe.super import SuperCond\n")
    f.write("from jarvis.db.jsonutils import loadjson\n")
    f.write('d=loadjson("' + str(job_json) + '")\n')
    f.write("v=SuperCond.from_dict(d)\n")
    f.write("v.runjob()\n")
    f.close()


jids = ["JVASP-816", "JVASP-19821"]
jids = [
    "POSCAR-AlN2Sc.vasp",
    "POSCAR-CrV6Pt.vasp",
    "POSCAR-NbNiV6.vasp",
    "POSCAR-RuNb.vasp",
]
submit_job = True
use_preconverged_kpoints = False
for i in jids:
    try:
        print("jid", i)
        if "POSCAR" not in i:
            dat = get_jid_data(jid=i, dataset="dft_3d")
            a_atoms = Atoms.from_dict(dat["atoms"])
        else:
            a_atoms = Atoms.from_poscar(i)
        dir_name = os.path.join(run_dir, i + "_SUPER")
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        os.chdir(dir_name)

        atoms = Spacegroup3D(a_atoms).refined_atoms.get_primitive_atoms
        # print (atoms)
        if use_preconverged_kpoints:

            kp = Kpoints3D().automatic_length_mesh(
                # lattice_mat=atoms.lattice_mat,
                # length=10
                lattice_mat=atoms.lattice_mat,
                length=dat["kpoint_length_unit"],
            )
            kpts = kp._kpoints[0]
            kpts = non_prime_kpoints(kpts)
            kp = Kpoints3D(kpoints=[kpts])
            print("kpts", kpts)

            nq1 = get_factors(kpts[0])[0]
            nq2 = get_factors(kpts[1])[0]
            nq3 = get_factors(kpts[2])[0]
            qp = Kpoints3D(kpoints=[[nq1, nq2, nq3]])
        else:
            kp = Kpoints3D(kpoints=[])
            qp = Kpoints3D(kpoints=[])
        sup = SuperCond(atoms=atoms, kp=kp, qp=qp, qe_cmd=qe_cmd).to_dict()
        dumpjson(data=sup, filename="sup.json")
        write_qejob(job_json=os.path.abspath("sup.json"))
        path = (
            "echo hello"
            + " \n module load intel/2015\nmodule load "
            + "openmpi/2.1.0/intel-15\nsource activate qe \npython "
            + os.getcwd()
            + "/job.py"
        )

        if submit_job:
            Queue.slurm(
                job_line=path,
                jobname=i,
                directory=os.getcwd(),
                # queue="mml",
                walltime="330:0:0",
                submit_cmd=["sbatch", "submit_job"],
            )

        os.chdir("..")
    except Exception as exp:
        print(exp)
        pass



"""Module for running QMCPack."""
from jarvis.io.nexus.inputs import (
    get_Zeff,
    get_pseudo_dft_dict,
    # get_pseudo_qmc_dict,
)
from jarvis.io.qe.outputs import QEout
from nexus import generate_physical_system
from nexus import generate_pwscf
from nexus import settings, job, run_project, obj
from jarvis.core.atoms import Atoms
from jarvis.db.figshare import get_jid_data
from nexus import read_structure
from jarvis.core.kpoints import Kpoints3D
import numpy as np
import os


def get_energy_nexus(
    atoms=None,
    poscar_path=None,
    length=10,
    ecut=40,
    temp_filename="POSCAR-temp.vasp",
    qe_presub="module load qe/7.0.pw2qmcpack",
    qmcpack_presub="module load qmcpack/3.14.0",
    nodes=2,
    cores=16,
    threads=1,
    app="pw.x",
    net_spin=0,
    path="scf",
    machine="dobby",
    pseudo_path="/rk2/knc6/QMC/pseudopotentials",
    results_path="/users/knc6/Software/jarvis/usnistgov/jarvis/results",
    runs_path="/users/knc6/Software/jarvis/usnistgov/jarvis/runs",
):
    """Get energy from Nexus+qmcpack."""
    results = results_path
    runs = runs_path
    settings(
        results=results,
        pseudo_dir=pseudo_path,  # location of pseudopotential directory
        sleep=1,
        runs=runs,
        machine=machine,
    )
    if atoms is None:
        atoms = Atoms.from_poscar(poscar_path)
        temp_filename = poscar_path

    kp = (
        Kpoints3D()
        .automatic_length_mesh(lattice_mat=atoms.lattice_mat, length=length)
        ._kpoints
    )
    kp = np.array(kp).flatten().tolist()
    atoms.write_poscar(temp_filename)
    structure = read_structure(temp_filename, format="poscar")
    print("structure", structure)
    print("Running job")
    shared_qe = obj(
        occupations="smearing",
        smearing="gaussian",
        degauss=0.005,
        input_dft="PBE",
        ecut=ecut,  # Plane Wave cutoff energy
        conv_thr=1.0e-7,
        mixing_beta=0.2,
        nosym=True,
        use_folded=True,
        spin_polarized=True,
    )

    qe_job = job(
        nodes=nodes, cores=cores, threads=threads, app=app, presub=qe_presub
    )
    Zeff = get_Zeff()
    system = generate_physical_system(
        structure=structure,
        kshift=(0, 0, 0),
        net_spin=net_spin,  # Specify Up - Down Electrons
        **{e: Zeff[e] for e in structure.elem}
    )
    pseudo_dft_dict = get_pseudo_dft_dict()
    arr = []
    for sp in atoms.elements:
        arr.append(pseudo_dft_dict[sp])
    scf = generate_pwscf(
        identifier="scf",  # log output goes to scf.out
        path="scf",  # directory to run in
        job=qe_job,  # pyscf must run w/o mpi
        system=system,
        input_type="scf",
        pseudos=arr,
        kgrid=kp,
        wf_collect=False,
        **shared_qe
    )
    print('scf', scf)
    run_project()
    out_file = os.path.join(runs, "scf", "scf.out")
    print("out_file", out_file)
    qeout = QEout(filename=out_file)
    tot_energy = qeout.get_total_energy()
    print("tot_energy", tot_energy)
    return tot_energy


if __name__ == "__main__":
    jid = "JVASP-32"
    atoms = Atoms.from_dict(
        get_jid_data(jid="JVASP-1002", dataset="dft_3d")["atoms"]
    )
    get_energy_nexus(atoms=atoms)



"""Get formation energy, convex hull etc.."""

import os
from scipy.spatial import ConvexHull
import numpy as np
from jarvis.db.figshare import data
from jarvis.core.atoms import Atoms
from jarvis.db.jsonutils import loadjson
from collections import OrderedDict
from jarvis.core.composition import Composition
import re

# import matplotlib.pyplot as plt


def get_optb88vdw_energy():
    """Get OptB88vdW energy per atoms for elements."""
    return loadjson(os.path.join(os.path.dirname(__file__), "unary.json"))


def get_unary_qe_tb_energy():
    """Get elemental chemical potential for GBRV tight-binding project."""
    return loadjson(
        os.path.join(os.path.dirname(__file__), "unary_qe_tb.json")
    )


def isfloat(value):
    """Check if a number is float.

    TODO: replace with isinstance.
    """
    try:
        float(value)
        return True
    except ValueError:
        return False


def unary_energy(el="Na", chem_pots=None):
    """Provide energy per atoms of an element."""
    if chem_pots is None:
        chem_pots = get_optb88vdw_energy()
    en = "na"
    for i, j in chem_pots.items():
        if str(i) == str(el):
            en = j["energy"]
    return en


def form_enp(atoms=None, total_energy=None, chem_pots=None):
    """
    Calculate formation energy given the total energy and the atoms object.

    Currently for OptB88vdW functional based chemical potential implemented
    but can be generalized by changing unary_energy.
    """
    dd = atoms.composition.to_dict()
    # print ('dd',dd)
    ref = 0.0
    for k, v in dd.items():
        e1 = unary_energy(el=k, chem_pots=chem_pots)
        # print (k,v,e1,total_energy)
        if e1 == "na":
            ref = "na"
            ValueError("Element reference does not exist", e1)

        else:
            ref = ref + float(v) * float(e1)
    if isfloat(ref):
        form_en = float(total_energy) - float(ref)
        form_en = round(float(form_en) / float(atoms.num_atoms), 5)
    return form_en


def get_twod_defect_energy(vrun="", jid="", atom=""):
    """Get mono 2D defect formation energy with OptB88vdW data."""
    dft2d = data("dft_2d")

    def get_enp_jid(jid=""):
        for i in dft2d:
            if i["jid"] == jid:
                return (
                    i["optb88vdw_total_energy"]
                    / Atoms.from_dict(i["atoms"]).num_atoms
                )

        # dir='JVASP-667_C_C_c'
        # tmp=dir.split('_')
        # jid=tmp[0]
        # atom=tmp[2]

    strt = vrun.all_structures[-1]
    natoms = strt.num_atoms
    fin_en = vrun.final_energy
    chem_pot = unary_energy(atom)
    bulk_en_pa = get_enp_jid(jid)
    Ef = fin_en - (natoms + 1) * bulk_en_pa + chem_pot
    return Ef


class PhaseDiagram:
    """Module for phase diagram."""

    def __init__(
        self,
        entries,
        verbose=False,
        only_plot_stable=False,
        only_label_stable=False,
    ):
        """Initialize Phase-diagram."""
        # Adapted from ASE
        self.species = OrderedDict()
        # List of formula,formation energy,JID etc.
        self.entries = entries
        self.entries_dict = []
        self.verbose = verbose
        self.only_plot_stable = only_plot_stable
        self.only_label_stable = only_label_stable
        for i in self.entries:
            name = i[0]
            energy = i[1]
            # jid = i[2]
            count = Composition.from_string(name).to_dict()

            natoms = 0
            for symbol, n in count.items():
                natoms += n
                if symbol not in self.species:
                    self.species[symbol] = len(self.species)
            self.entries_dict.append((count, energy, name, natoms))

        ns = len(self.species)
        self.symbols = [None] * ns
        for symbol, id in self.species.items():
            self.symbols[id] = symbol

        if verbose:
            print("Species:", ", ".join(self.symbols))
            print("Entries:", len(self.entries_dict))
            for i, (count, energy, name, natoms) in enumerate(
                self.entries_dict
            ):
                print("{:<5}{:10}{:10.3f}".format(i, name, energy))

        self.points = np.zeros((len(self.entries_dict), ns + 1))
        for s, (count, energy, name, natoms) in enumerate(self.entries_dict):
            for symbol, n in count.items():
                self.points[s, self.species[symbol]] = n / natoms
            self.points[s, -1] = energy  # / natoms

        if len(self.points) == ns:
            # Simple case that qhull would choke on:
            self.simplices = np.arange(ns).reshape((1, ns))
            self.hull = np.ones(ns, bool)
        else:
            # print("self.points[:, 1:]",self.points[:, 1:])
            hull = ConvexHull(self.points[:, 1:])

            # Find relevant simplices:
            ok = hull.equations[:, -2] < 0
            self.simplices = hull.simplices[ok]

            # Create a mask for those points that are on the convex hull:
            self.hull = np.zeros(len(self.points), bool)
            for simplex in self.simplices:
                self.hull[simplex] = True

    def energy_above_hull(self, entry=[]):
        """Find energy above hull."""
        formula = entry[0]
        form_enp = entry[1]
        kwargs = Composition.from_string(formula).to_dict()

        point = np.zeros(len(self.species))
        N = 0
        for symbol, n in kwargs.items():
            point[self.species[symbol]] = n
            N += n
        # print ('N',N)
        # Find coordinates within each simplex:
        X = self.points[self.simplices, 1:-1] - point[1:] / N

        # Find the simplex with positive coordinates that sum to
        # less than one:
        eps = 1e-15
        for i, Y in enumerate(X):
            try:
                x = np.linalg.solve((Y[1:] - Y[:1]).T, -Y[0])
            except np.linalg.linalg.LinAlgError:
                continue
            if (x > -eps).all() and x.sum() < 1 + eps:
                break
        else:
            assert False, X

        indices = self.simplices[i]
        points = self.points[indices]

        scaledcoefs = [1 - x.sum()]
        scaledcoefs.extend(x)
        #         print('scaledcoefs',scaledcoefs)
        #         print('points[:, -1]',points[:, -1])
        energy = np.dot(scaledcoefs, points[:, -1])  # *N

        coefs = []
        results = []
        for coef, s in zip(scaledcoefs, indices):
            count, e, name, natoms = self.entries_dict[s]
            coef *= N / natoms
            coefs.append(coef)
            results.append((name, coef, e))

        #         if self.verbose:
        #             print_results(results)
        e_above_hull = form_enp - energy
        return e_above_hull, energy, indices, np.array(coefs)

    def get_ehull_all(self):
        """Find energy above hull for all entries."""
        info = []
        for i in self.entries:
            # print('ent',i)
            ehull, energy, indices, coefs = self.energy_above_hull(
                entry=[i[0], i[1]]
            )
            info.append([i, ehull])
        return info

    def plot(self, ax=None, dims=None, show=False):
        """Make 2-d or 3-d plot of datapoints and convex hull.

        Default is 2-d for 2- and 3-component diagrams and 3-d for a
        4-component diagram.
        """
        import matplotlib.pyplot as plt

        N = len(self.species)

        if dims is None:
            if N <= 3:
                dims = 2
            else:
                dims = 3

        if ax is None:
            projection = None
            if dims == 3:
                projection = "3d"
                from mpl_toolkits.mplot3d import Axes3D

                Axes3D  # silence pyflakes
            fig = plt.figure()
            ax = fig.add_subplot(projection=projection)
        else:
            if dims == 3 and not hasattr(ax, "set_zlim"):
                raise ValueError(
                    "Cannot make 3d plot unless axes projection " "is 3d"
                )

        if dims == 2:
            if N == 2:
                self.plot2d2(ax)
            elif N == 3:
                self.plot2d3(ax)
            else:
                raise ValueError(
                    "Can only make 2-d plots for 2 and 3 " "component systems!"
                )
        else:
            if N == 3:
                self.plot3d3(ax)
            elif N == 4:
                self.plot3d4(ax)
            else:
                raise ValueError(
                    "Can only make 3-d plots for 3 and 4 " "component systems!"
                )

        if show:
            plt.show()
        return ax

    def plot2d2(self, ax=None):
        """Get 2D plot."""
        x, e = self.points[:, 1:].T
        names = [
            re.sub(r"(\d+)", r"$_{\1}$", ref[2]) for ref in self.entries_dict
        ]
        hull = self.hull
        simplices = self.simplices
        xlabel = self.symbols[1]
        ylabel = "energy [eV/atom]"
        extra = -min(e) / 10
        if ax:
            for i, j in simplices:
                ax.plot(x[[i, j]], e[[i, j]], "-b")
            ax.plot(x[hull], e[hull], "sg")
            if not self.only_plot_stable:
                ax.plot(x[~hull], e[~hull], "or")

            if self.only_plot_stable or self.only_label_stable:
                x = x[self.hull]
                e = e[self.hull]
                names = [name for name, h in zip(names, self.hull) if h]
            for a, b, name in zip(x, e, names):
                if b <= extra:
                    ax.text(a, b, name, ha="center", va="top")

            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            ax.set_ylim([min(e) - extra, extra])
        return (x, e, names, hull, simplices, xlabel, ylabel)

    def plot2d3(self, ax=None):
        """Get 2D plot for ternaries."""
        x, y = self.points[:, 1:-1].T.copy()
        x += y / 2
        y *= 3 ** 0.5 / 2
        names = [
            re.sub(r"(\d+)", r"$_{\1}$", ref[2]) for ref in self.entries_dict
        ]
        hull = self.hull
        simplices = self.simplices

        if ax:
            for i, j, k in simplices:
                ax.plot(x[[i, j, k, i]], y[[i, j, k, i]], "-b")
            ax.plot(x[hull], y[hull], "og")
            if not self.only_plot_stable:
                ax.plot(x[~hull], y[~hull], "sr")
            if self.only_plot_stable or self.only_label_stable:
                x = x[self.hull]
                y = y[self.hull]
                names = [name for name, h in zip(names, self.hull) if h]
            for a, b, name in zip(x, y, names):
                ax.text(a, b, name, ha="center", va="top")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")
        return (x, y, names, hull, simplices)

    def plot3d3(self, ax):
        """Get 3D plot for ternaries."""
        x, y, e = self.points[:, 1:].T

        ax.scatter(x[self.hull], y[self.hull], e[self.hull], c="g", marker="o")
        if not self.only_plot_stable:
            ax.scatter(
                x[~self.hull], y[~self.hull], e[~self.hull], c="r", marker="s"
            )

        for a, b, c, ref in zip(x, y, e, self.entries_dict):
            name = re.sub(r"(\d+)", r"$_{\1}$", ref[2])
            ax.text(a, b, c, name, ha="center", va="bottom")

        for i, j, k in self.simplices:
            ax.plot(
                x[[i, j, k, i]], y[[i, j, k, i]], zs=e[[i, j, k, i]], c="b"
            )

        ax.set_xlim3d(0, 1)
        ax.set_ylim3d(0, 1)
        ax.view_init(azim=115, elev=30)
        ax.set_xlabel(self.symbols[1])
        ax.set_ylabel(self.symbols[2])
        ax.set_zlabel("energy [eV/atom]")
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")

    def plot3d4(self, ax):
        """Get 3D plot for quaternaries."""
        x, y, z = self.points[:, 1:-1].T
        a = x / 2 + y + z / 2
        b = 3 ** 0.5 * (x / 2 + y / 6)
        c = (2 / 3) ** 0.5 * z

        ax.scatter(a[self.hull], b[self.hull], c[self.hull], c="g", marker="o")
        if not self.only_plot_stable:
            ax.scatter(
                a[~self.hull], b[~self.hull], c[~self.hull], c="r", marker="s"
            )

        for x, y, z, ref in zip(a, b, c, self.entries_dict):
            name = re.sub(r"(\d+)", r"$_{\1}$", ref[2])
            ax.text(x, y, z, name, ha="center", va="bottom")

        for i, j, k, w in self.simplices:
            ax.plot(
                a[[i, j, k, i, w, k, j, w]],
                b[[i, j, k, i, w, k, j, w]],
                zs=c[[i, j, k, i, w, k, j, w]],
                c="b",
            )

        ax.set_xlim3d(0, 1)
        ax.set_ylim3d(0, 1)
        ax.set_zlim3d(0, 1)
        ax.view_init(azim=115, elev=30)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")


def jid_hull(jid="", dataset=[]):
    """Get ehull for a jid and a dataset e.g. dft_3d."""
    from jarvis.db.figshare import data

    if isinstance(dataset, str):
        dataset = data(dataset)
    for i in dataset:
        if i["jid"] == jid:
            system = list(set(i["atoms"]["elements"]))
    z = []
    for i in dataset:
        formula = i["formula"]
        comp = Composition.from_string(formula)
        # atom_frac = comp.atomic_fraction
        all_elms = list(comp.to_dict())
        if (set(all_elms)).issubset(set(system)):
            z.append([i["formula"], i["formation_energy_peratom"], i["jid"]])

    pdj = PhaseDiagram(z)
    # pdj.plot()
    info = pdj.get_ehull_all()
    for i in info:
        if i[0][2] == jid:
            return i


def formula_hull(formula_energy_id=[], dataset=[]):
    """Get ehull for a formula_energy_id pair and a dataset e.g. dft_3d."""
    # e.g. ["Al2O3",-1.0,"JVASP-xyz"]
    # for i in dataset:
    #    if i["jid"] == jid:
    #        system = list(set(i["atoms"]["elements"]))
    from jarvis.db.figshare import data

    if isinstance(dataset, str):
        dataset = data(dataset)

    c = Composition.from_string(formula_energy_id[0])
    system = list(c.to_dict().keys())
    z = []
    z.append(formula_energy_id)

    for i in dataset:
        formula = i["formula"]
        comp = Composition.from_string(formula)
        # atom_frac = comp.atomic_fraction
        all_elms = list(comp.to_dict())
        if (set(all_elms)).issubset(set(system)):
            z.append([i["formula"], i["formation_energy_peratom"], i["jid"]])

    pdj = PhaseDiagram(z)
    # pdj.plot()
    info = pdj.get_ehull_all()
    for i in info:
        if i[0][2] == formula_energy_id[-1]:
            return i



"""Module to get elemnt fraction based array for a formula."""
# Maximum upto 103 elements
import numpy as np
from jarvis.core.composition import Composition
from jarvis.core.specie import Specie


def get_element_fraction_desc(formula="SiO2", max_nelements=103):
    """Get element fraction."""
    x = np.zeros(max_nelements)
    fracs = Composition.from_string(formula).atomic_fraction
    for i, j in fracs.items():
        # -1 because python array srats from 0
        x[Specie(i).Z - 1] = j
    return x



"""Module for processing force-constant."""
import numpy as np

"""
def dynamical_matrix(
    force_constant=[],
    atoms=None,
    nac_params={},
    num_G_points=300,
    g_rad=100,
    exp_cutoff=1e-10,
):
    G_cutoff = (3 * num_G_points / (4 * np.pi) / atoms.volume) ** (1.0 / 3)
    rec_lat = atoms.lattice.inv_lattice()
    pts = np.arange(-g_rad, g_rad + 1)
    grid = np.meshgrid(pts, pts, pts)
    for i in range(3):
        grid[i] = grid[i].ravel()
    G_vec_list = np.dot(rec_lat, grid).T
    G_norm2 = ((G_vec_list) ** 2).sum(axis=1)
    G_list = np.array(
        G_vec_list[G_norm2 < G_cutoff ** 2], dtype="double", order="C"
    )
    GeG = G_cutoff ** 2 * np.trace(nac_params["epsilon"]) / 3
    Lambda = np.sqrt(-GeG / 4 / np.log(exp_cutoff))
"""


def read_fc(filename="FORCE_CONSTANTS"):
    """Read force-constants."""
    from jarvis.io.phonopy.outputs import read_fc

    fc = read_fc(filename)
    return fc


def qpoint(force_constant=[], qpt=[0.0, 0.0, 0.0]):
    """Get FC as natons x natons x 3 x3."""
    qpt = np.array(qpt)
    exp_iqpt = np.exp(1.0j * qpt)
    dmat = force_constant * exp_iqpt
    vals, vects = np.linalg.eigh(dmat)
    return vals, vects



from jarvis.core.atoms import Atoms
from jarvis.core.graphs import (
    StructureDataset,
    Graph,
    prepare_dgl_batch,
    prepare_line_graph_batch,
)
from jarvis.db.figshare import data
import pandas as pd
import os
import torch

test_pos = os.path.join(
    os.path.dirname(__file__),
    "POSCAR-JVASP-7577",
)


def test_graph():
    from jarvis.core.atoms import Atoms
    from jarvis.db.figshare import get_jid_data

    atoms = Atoms.from_poscar(test_pos)
    g = Graph.atom_dgl_multigraph(atoms=atoms, atom_features="cgcnn")
    atoms = Atoms.from_dict(get_jid_data("JVASP-664")["atoms"])
    feature_sets = ("atomic_number", "basic", "cfid", "cgcnn")
    for i in feature_sets:
        g = Graph.atom_dgl_multigraph(atoms=atoms, atom_features=i)
        g = Graph.atom_dgl_multigraph(atoms=atoms, atom_features=i)
        print(i, g)
    batch = prepare_dgl_batch(torch.tensor([1, 1]))
    batch = prepare_line_graph_batch(torch.tensor([1, 1, 1]))
    g = Graph.from_atoms(atoms=atoms, features="atomic_number")
    gnx = g.to_networkx()
    g = Graph.from_atoms(atoms=atoms, features="atomic_number")
    g = Graph.from_atoms(atoms=atoms, features="atomic_fraction")
    g = Graph.from_atoms(
        atoms=atoms,
        features="basic",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(
        atoms=atoms,
        features="cfid",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(
        atoms=atoms,
        features="atomic_number",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(atoms=atoms, features="basic")
    g = Graph.from_atoms(
        atoms=atoms, features=["Z", "atom_mass", "max_oxid_s"]
    )
    g = Graph.from_atoms(atoms=atoms, features="cfid", max_cut=10000)
    print(g)
    d = g.to_dict()
    g = Graph.from_dict(d)
    num_nodes = g.num_nodes
    num_edges = g.num_edges
    print(num_nodes, num_edges)
    assert num_nodes == 48
    assert num_edges == 2256
    assert (g.adjacency_matrix.shape) == (48, 48)


def test_dataset():
    d = data("dft_2d")
    d = pd.DataFrame(d[:100])

    graphs = []
    for i, row in d.iterrows():
        structure = Atoms.from_dict(row["atoms"])
        g = Graph.atom_dgl_multigraph(
            structure,
            atom_features="atomic_number",
            compute_line_graph=False,
        )
        graphs.append(g)

    s = StructureDataset(
        d, graphs, "formation_energy_peratom", line_graph=True
    )
    col = s.collate
    col1 = s.collate_line_graph
    ix = s[0]
    sz = len(s)
    s_std = s.setup_standardizer([1, 2])


# test_graph()



from jarvis.analysis.defects.substitutions import generate_defect
from jarvis.core.atoms import Atoms
from jarvis.io.vasp.inputs import Poscar
import os


def test_subs():
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Al", "Al"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    v = generate_defect(atoms=Si)


# test_2d()
# test_vacancy()



"""Modules for generating input files for QE."""

import os
import requests
import tarfile
from jarvis.core.specie import Specie
import numpy as np

# from jarvis.analysis.structure.spacegroup import Spacegroup3D


class QEinfile(object):
    """
    Gnererate quantum espresso input files.

    jarvis.core.Atoms and Kpoints are required as inputs.
    For other input parameters,
    see https://www.quantum-espresso.org/Doc/INPUT_PW.html .
    """

    def __init__(
        self,
        atoms=None,
        kpoints=None,
        psp_dir=None,
        input_params={},
        url=None,
        sanitize=True,
        sanitize_tol=2e-4,
        psp_temp_name=None,
    ):
        """Initialize input parameters for qunatum espresso."""
        if psp_temp_name is None:
            psp_temp_name = "QE_PSPs"
        if input_params == {}:
            input_params = GenericInputs().geometry_optimization()
        if psp_dir is None:
            psp_dir = str(
                os.path.join(os.path.dirname(__file__), psp_temp_name)
            )
            # Download GBRV PSPs by default
            if url is None:
                url = (
                    "https://www.physics.rutgers.edu/"
                    "gbrv/all_pbesol_UPF_v1.5.tar.gz"
                )
                print("Please cite for PSPs:")
                print("https://doi.org/10.1016/j.commatsci.2013.08.053")
            if not os.path.exists(psp_dir):
                print("Downloading PSPs")
                tar_file_name = str(
                    os.path.join(os.path.dirname(__file__), "psp.tar.gz")
                )
                r = requests.get(url)
                f = open(tar_file_name, "wb")
                f.write(r.content)
                f.close()
                tar = tarfile.open(tar_file_name)
                tar.extractall(psp_dir)
                tar.close()
                os.remove(tar_file_name)
        self.species = []
        self.psp_dir = psp_dir
        self.input_params = input_params
        self.atoms = atoms
        self.kpoints = kpoints
        self.sanitize = sanitize
        self.sanitize_tol = sanitize_tol
        if "system" in input_params:
            self.system_params = input_params["system"]
            if (
                input_params["system"]["nat"] is None
                or input_params["system"]["nat"] == ""
            ):
                input_params["system"]["nat"] = self.atoms.num_atoms
            if (
                input_params["system"]["ntyp"] is None
                or input_params["system"]["ntyp"] == ""
            ):
                self.species = self.atoms.uniq_species
                input_params["system"]["ntyp"] = len(self.species)
            if (
                "nspin" in input_params["system"]
                and input_params["system"]["nspin"] == 2
            ):

                for ii in range(input_params["system"]["nat"]):
                    tmp = "starting_magnetization(" + str(ii + 1) + str(")")
                    input_params["system"][tmp] = 1.0

                # 'starting_magnetization' in input_params["system"] and
                # input_params["system"]['starting_magnetization'] is None:

        else:
            self.system_params = {}

        if "electrons" in input_params:
            self.electron_params = input_params["electrons"]
        else:
            self.electron_params = {}

        if "control" in input_params:
            self.control_params = input_params["control"]
            self.control_params["pseudo_dir"] = str("'") + self.psp_dir + "'"
            if (
                "prefix" in self.control_params
                and self.control_params["prefix"] is None
            ):
                self.control_params[
                    "prefix"
                ] = self.atoms.composition.reduced_formula
        else:
            self.control_params = {}

        if "inputa2f" in input_params:
            self.inputa2f = input_params["inputa2f"]
        else:
            self.inputa2f = {}

        if "ions" in input_params:
            self.ion_params = input_params["ions"]
        else:
            self.ion_params = {}

        if "cell" in input_params:
            self.cell_params = input_params["cell"]
        else:
            self.cell_params = {}
        if "input" in input_params:
            self.input = input_params["input"]
            if (
                "amass(1)" not in input_params["input"]
                and "zasr" not in input_params["input"]
            ):
                for ii, jj in enumerate(self.atoms.uniq_species):
                    tmp = "amass(" + str(ii + 1) + ")"
                    input_params["input"][tmp] = str(
                        round(Specie(jj).atomic_mass, 2)
                    )

        else:
            self.input = {}

        if "inputph" in input_params:
            self.inputph = input_params["inputph"]
            if "amass(1)" not in input_params["inputph"]:
                for ii, jj in enumerate(self.atoms.uniq_species):
                    tmp = "amass(" + str(ii + 1) + ")"
                    input_params["inputph"][tmp] = str(
                        round(Specie(jj).atomic_mass, 2)
                    )
        else:
            self.inputph = {}

    def dictionary_to_string(self, tags={}):
        """Convert a dictionary to string with '=' sign."""
        lines = ""
        for i, j in tags.items():
            if j is not None:
                lines = lines + str(i) + str(" = ") + str(j) + "\n"
        return lines

    def kpoints_to_string(self):
        """Convert a jarvis.core.Kpoints3D to string."""
        kp = ""
        if self.kpoints:
            kpoint_mode = self.kpoints._kpoint_mode
            if kpoint_mode == "automatic":
                kp = (
                    kp
                    + "K_POINTS automatic\n"
                    + (" ".join(map(str, self.kpoints.kpts[0])) + " 0 0 0\n")
                )
            elif kpoint_mode == "linemode":
                points = ""
                for i in self.kpoints.kpts:
                    points = points + " ".join(map(str, i)) + " 1.0" + "\n"
                kp = (
                    kp
                    + "K_POINTS crystal\n"
                    + str(len(self.kpoints.kpts))
                    + "\n"
                    + points
                )

            else:
                print(
                    "Kpoint scheme not implemented except linemode&automatic"
                )
        return kp

    def get_psp(self, element):
        """Obtain psuedopotential for an element."""
        element = str(element).lower()
        for i in os.listdir(self.psp_dir):
            el = str(i.split(".")[0].split("_")[0]).lower()
            if el == element:
                return i

    def atomic_species_string(self):
        """Obtain string for QE atomic species."""
        line = ""
        for i in self.species:
            line = line + (
                i
                + " "
                + str(round(Specie(i).atomic_mass, 2))
                + " "
                + str(self.get_psp(i))
                + "\n"
            )
        return line

    def check_frac(self, n):
        """Check fractional coordinates or lattice parameters."""
        items = [
            0.0,
            0.3333333333333333,
            0.25,
            0.5,
            0.75,
            0.6666666666666667,
            1.0,
            1.5,
            2.0,
            -0.5,
            -2.0,
            -1.5,
            -1.0,
            1.0 / 2.0 ** 0.5,
            -1.0 / 2.0 ** 0.5,
            3.0 ** 0.5 / 2.0,
            -(3.0 ** 0.5) / 2.0,
            1.0 / 3.0 ** 0.5,
            -1.0 / 3.0 ** 0.5,
            1.0 / 2.0 / 3 ** 0.5,
            -1.0 / 2.0 / 3 ** 0.5,
            1 / 6,
            5 / 6,
        ]
        items = items + [(-1) * i for i in items]
        for f in items:
            if abs(f - n) < self.sanitize_tol:
                return f
        return n

    def atomic_pos(self):
        """Obtain string for QE atomic positions."""
        line = ""
        # self.atoms = Spacegroup3D(self.atoms).refined_atoms
        coords = np.array(self.atoms.frac_coords)
        ntot = self.atoms.num_atoms

        if self.sanitize:
            for i in range(ntot):
                for j in range(3):  # neatin
                    coords[i, j] = self.check_frac(coords[i, j])

        for i, j in zip(self.atoms.elements, coords):
            line = line + str(i) + " " + " ".join(map(str, j)) + "\n"
        return line

    def atomic_cell_params(self):
        """Obtain string for QE atomic lattice parameters."""
        lat_mat = np.array(self.atoms.lattice_mat)
        if self.sanitize:
            print("Sanitizing Atoms.")
            a_lat = np.linalg.norm(lat_mat[0, :])
            at = lat_mat / a_lat
            for i in range(3):
                for j in range(3):
                    at[i, j] = self.check_frac(at[i, j])

            lat_mat = at * a_lat

        line = (
            str(lat_mat[0][0])
            + " "
            + str(lat_mat[0][1])
            + " "
            + str(lat_mat[0][2])
            + "\n"
            + str(lat_mat[1][0])
            + " "
            + str(lat_mat[1][1])
            + " "
            + str(lat_mat[1][2])
            + "\n"
            + str(lat_mat[2][0])
            + " "
            + str(lat_mat[2][1])
            + " "
            + str(lat_mat[2][2])
        )
        return line

    def to_string(self):
        """Convert inputs to a string to write in file."""
        control = ""
        system = ""
        electrons = ""
        ions = ""
        cell = ""
        input = ""
        inputph = ""
        inputa2f = ""
        spec = ""
        if self.control_params:
            control = (
                "&control\n\n"
                + self.dictionary_to_string(self.control_params)
                + "/"
                + "\n"
            )
        if self.system_params:
            system = (
                "\n&system\n\n"
                + self.dictionary_to_string(self.system_params)
                + "/"
                + "\n"
            )
        if self.electron_params:
            electrons = (
                "\n&electrons\n\n"
                + self.dictionary_to_string(self.electron_params)
                + "/"
                + "\n"
            )
        if self.ion_params:
            ions = (
                "\n&ions\n\n"
                + self.dictionary_to_string(self.ion_params)
                + "/"
                + "\n"
            )
        if self.cell_params:
            cell = (
                "\n&cell\n\n"
                + self.dictionary_to_string(self.cell_params)
                + "/"
                + "\n"
            )
        if self.input:
            input = (
                "\n&input\n\n"
                + self.dictionary_to_string(self.input)
                + "/"
                + "\n"
            )
        if self.inputph:
            inputph = (
                "\n&inputph\n\n"
                + self.dictionary_to_string(self.inputph)
                + "/"
                + "\n"
            )
        if self.inputa2f:
            inputa2f = (
                "\n&inputa2F\n\n"
                + self.dictionary_to_string(self.inputa2f)
                + "/"
                + "\n"
            )
        if self.species:
            spec = "ATOMIC_SPECIES\n\n" + self.atomic_species_string() + "\n"
        line = (
            control
            + system
            + electrons
            + ions
            + cell
            + input
            + inputph
            + inputa2f
            + spec
            # + "ATOMIC_SPECIES\n\n"
            # + self.atomic_species_string()
            # + "\n"
            + "ATOMIC_POSITIONS crystal\n\n"
            + self.atomic_pos()
            + "\n"
            + "CELL_PARAMETERS angstrom\n\n"
            + self.atomic_cell_params()
            + "\n\n"
            + self.kpoints_to_string()
        )
        return line

    def write_file(self, filename="qe.in"):
        """Write input file."""
        line = self.to_string()
        f = open(filename, "w")
        f.write(line)
        f.close()


class GenericInputs(object):
    """Obtain set of minimal QE input parameters."""

    def __init__(self):
        """Initialize with minimum parameters that can be updated."""
        self.sample_qe_inputs = {
            "control": {
                "calculation": "'scf'",
                "restart_mode": "'from_scratch'",
                "prefix": "'QE'",
                "outdir": "'./'",
                "tstress": ".true.",
                "tprnfor": ".true.",
                "disk_io": "'low'",
                "wf_collect": ".true.",
                "pseudo_dir": None,
                "verbosity": "'high'",
                "nstep": 100,
            },
            "system": {
                "ibrav": 0,
                "nat": None,
                "ntyp": None,
                "ecutwfc": 45,
                "ecutrho": 250,
                "q2sigma": 1,
                "ecfixed": 44.5,
                "qcutz": 800,
                "occupations": "'smearing'",
                "degauss": 0.01,
                "lda_plus_u": ".false.",
                "force_symmorphic": ".true.",
                "nosym": ".false.",
                "noinv": ".false.",
            },
            "electrons": {
                "diagonalization": "'david'",
                "mixing_mode": "'local-TF'",
                "mixing_beta": 0.3,
                "conv_thr": "1d-9",
            },
            "ions": {"ion_dynamics": "'bfgs'"},
            "cell": {
                "cell_dynamics": "'bfgs'",
                "cell_dofree": "'all'",
                "cell_factor": 1.0,
            },
        }

    def geometry_optimization(self):
        """Obtain QE inputs for geometry optimization."""
        input = self.sample_qe_inputs
        input["control"]["calculation"] = "'vc-relax'"
        return input

    # def simple_scf(self):
    #    """Obtain QE inputs for single SCF."""
    #    input = self.sample_qe_inputs
    #    return input

    # def electron_band_structure(self):
    #    """Obtain QE inputs for bandstructure calculations optimization."""
    #    input = self.sample_qe_inputs
    #    # TODO: increase nbands
    #    input["control_params"]["calculation"] = "'nscf'"
    #    input["control_params"]["restart_mode"] = None
    #    return input


"""
if __name__ == "__main__":
    from jarvis.core.kpoints import Kpoints3D
    from jarvis.core.atoms import Atoms
    box = [[2.715, 2.715, 0], [0, 2.715, 2.715], [2.715, 0, 2.715]]
    coords = [[0, 0, 0], [0.25, 0.25, 0.25]]
    elements = ["Si", "Si"]
    Si = Atoms(lattice_mat=box, coords=coords, elements=elements)
    print(Si)
    kp = Kpoints3D().automatic_length_mesh(
        lattice_mat=Si.lattice_mat, length=20
    )
    qe = QEinfile(Si, kp)
    qe.write_file()
    kp = Kpoints3D().kpath(atoms=Si)
    qe = QEinfile(Si, kp)
    qe.write_file("qe.in2")
    sp = qe.atomic_species_string()
    sp = qe.atomic_cell_params()
    print("sp", sp)
    print(qe.input_params['system_params']['nat'])
#  /users/knc6/Software/QE/q-e/bin/pw.x -i qe.in
"""



"""
Modules for analyzing infrared intensities.

Please find more details in https://doi.org/10.1038/s41524-020-0337-2 .
"""
from jarvis.core.spectrum import Spectrum
import numpy as np
import os


def normalize_vecs(phonon_eigenvectors, masses):
    """
    Return the eigenvectors after division of each component by sqrt(mass).

    Adapted from https://github.com/JMSkelton/Phonopy-Spectroscopy/
    TODO: include LO-TO splitting.
    """
    nmodes = len(phonon_eigenvectors)
    nmasses = len(masses)
    natoms = nmasses
    sqrt_masses = np.sqrt(masses)
    eigendisplacements = np.zeros((nmodes, natoms, 3), dtype="float64")
    for i in range(0, nmodes):
        eigenvector = phonon_eigenvectors[i]
        for j in range(0, natoms):
            eigendisplacements[i, j, :] = np.divide(
                eigenvector[j], sqrt_masses[j]
            )
    return eigendisplacements


def ir_intensity(
    phonon_eigenvectors=[],
    phonon_eigenvalues=[],
    masses=[],
    born_charges=[],
    factor=33.35641,
    nac=True,
    epsilon=[],
    enforce_positive_freqs=True,
    smoothen=True,
):
    """Calculate IR intensity using DFPT."""
    # TODO:add non-analytical correction
    eigendisplacements = normalize_vecs(phonon_eigenvectors, masses)
    becDim1, becDim2, becDim3 = np.shape(born_charges)
    freq = []
    ir_ints = []
    for i, v in zip(eigendisplacements, phonon_eigenvalues):
        eigendisplacement = i
        eigDim1, eigDim2 = np.shape(eigendisplacement)
        irIntensity = 0.0
        for a in range(0, 3):
            sumTemp1 = 0.0
            for j in range(0, eigDim1):
                sumTemp2 = 0.0
                for b in range(0, 3):
                    sumTemp2 += born_charges[j][a][b] * eigendisplacement[j][b]
                sumTemp1 += sumTemp2
            irIntensity += sumTemp1 ** 2
        if enforce_positive_freqs:
            if v > 0:
                freq.append(v * factor)  # Thz to cm-1
                ir_ints.append(irIntensity)
        else:
            freq.append(v * factor)  # Thz to cm-1
            ir_ints.append(irIntensity)
    if smoothen:
        freq, ir_ints = Spectrum(x=freq, y=ir_ints).smoothen_spiky_spectrum()
    return freq, ir_ints


def ir_intensity_phonopy(
    run_dir=".",
    vasprun="vasprun.xml",
    BornFileName="BORN",
    PoscarName="POSCAR",
    ForceConstantsName="FORCE_CONSTANTS",
    supercell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    nac=True,
    symprec=1e-5,
    primitive=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
    degeneracy_tolerance=1e-5,
    vector=[0, 0, 0],
    smoothen=False,
):
    """Calculate IR intensity using DFPT and phonopy."""
    from phonopy import Phonopy
    from phonopy.interface.vasp import read_vasp
    from phonopy.file_IO import (
        parse_BORN,
        parse_FORCE_CONSTANTS,
    )
    import shutil
    from phonopy.units import VaspToCm

    # from phonopy.phonon.degeneracy import (
    #    degenerate_sets as get_degenerate_sets,
    # )
    # adapted from https://github.com/JaGeo/IR
    # TODO: Make directory indepndent
    cwd = os.getcwd()
    print("Directory:", cwd)
    os.chdir(run_dir)
    if not os.path.exists(vasprun):
        shutil.copy2(vasprun, "vasprun.xml")
    cmd = str("phonopy --fc vasprun.xml")
    os.system(cmd)
    born_file = os.path.join(os.getcwd(), BornFileName)
    cmd = str("phonopy-vasp-born >  ") + str(born_file)
    os.system(cmd)
    from jarvis.io.vasp.outputs import Vasprun

    v = Vasprun(vasprun)
    strt = v.all_structures[0]
    strt.write_poscar(PoscarName)
    unitcell = read_vasp(PoscarName)
    phonon = Phonopy(
        unitcell,
        supercell_matrix=supercell,
        primitive_matrix=primitive,
        factor=VaspToCm,
        symprec=symprec,
    )
    natoms = phonon.get_primitive().get_number_of_atoms()
    force_constants = parse_FORCE_CONSTANTS(filename=ForceConstantsName)
    phonon.set_force_constants(force_constants)
    masses = phonon.get_primitive().get_masses()
    phonon.set_masses(masses)
    BORN_file = parse_BORN(phonon.get_primitive(), filename=BornFileName)
    BORN_CHARGES = BORN_file["born"]
    # print ('born_charges2',BORN_CHARGES)
    if nac:
        phonon.set_nac_params(BORN_file)
    frequencies, eigvecs = phonon.get_frequencies_with_eigenvectors(vector)
    # frequencies=VaspToTHz*frequencies/VaspToCm
    # x, y = ir_intensity(
    #    phonon_eigenvectors=np.real(eigvecs),
    #    phonon_eigenvalues=frequencies,
    #    masses=masses, #np.ones(len(masses)),
    #    born_charges=born_charges,
    #    smoothen=smoothen,
    # )
    NumberOfBands = len(frequencies)
    EigFormat = {}
    for alpha in range(NumberOfBands):
        laufer = 0
        for beta in range(natoms):
            for xyz in range(0, 3):
                EigFormat[beta, alpha, xyz] = eigvecs[laufer][alpha]
                laufer = laufer + 1
    Intensity = {}
    intensities = []
    for freq in range(len(frequencies)):
        Intensity[freq] = 0
        tmp = 0
        for alpha in range(3):
            asum = 0
            for n in range(natoms):
                for beta in range(3):
                    asum = asum + BORN_CHARGES[n, alpha, beta] * np.real(
                        EigFormat[n, freq, beta]
                    ) / np.sqrt(masses[n])
                    tmp += asum
            Intensity[freq] = Intensity[freq] + np.power(np.absolute(asum), 2)
        intensities.append(Intensity[freq])
    os.chdir(cwd)
    return frequencies, intensities


"""
if __name__ == "__main__":
    from jarvis.io.vasp.outputs import Vasprun, Outcar

    out = Outcar("../../tests/testfiles/io/vasp/OUTCAR.JVASP-39")
    vrun = Vasprun("../../tests/testfiles/io/vasp/vasprun.xml.JVASP-39")
    phonon_eigenvectors = vrun.dfpt_data["phonon_eigenvectors"]
    vrun_eigs = vrun.dfpt_data["phonon_eigenvalues"]
    phonon_eigenvalues = out.phonon_eigenvalues
    masses = vrun.dfpt_data["masses"]
    born_charges = vrun.dfpt_data["born_charges"]
    x, y = ir_intensity(
        phonon_eigenvectors=phonon_eigenvectors,
        phonon_eigenvalues=phonon_eigenvalues,
        masses=masses,
        born_charges=born_charges,
        smoothen=False,
    )
    for i, j in zip(x, y):
        if j > 0.1:
            print("first", i, j)
    print()
    x, y = ir_intensity_phonopy()
    for i, j in zip(x, y):
        if j > 0.1:
            print("later", i, j)
    # for i, j in zip(phonon_eigenvalues, vrun_eigs):
    #    print(i, j)
"""



"""Module to prepare input file for prismatic."""

# Use jarvis.core.atoms.crop_squre for non rectangles


def write_prismatic_xyz(atoms=None):
    """Write Atoms in prismatic format."""
    from jarvis.core.specie import Specie

    line = str(atoms.num_atoms) + "\n"
    abc = atoms.lattice.abc
    line += str(abc[0]) + " " + str(abc[1]) + " " + str(abc[2]) + "\n"
    for i, j in zip(atoms.elements, atoms.cart_coords):
        line += (
            str(Specie(i).Z)
            + " "
            + str(j[0])
            + " "
            + str(j[1])
            + " "
            + str(j[2])
            + str("  1 .076\n")
        )
    line += "-1\n"
    return line



from jarvis.core.circuits import QuantumCircuitLibrary


def test_circuits():
    Q = QuantumCircuitLibrary()
    circs = [
        Q.circuit1(),
        Q.circuit2(),
        Q.circuit3(),
        Q.circuit4(),
        Q.circuit5(),
        Q.circuit6(),
    ]



"""Module to design quantum circuits with qiskit."""
from qiskit.circuit import QuantumCircuit, ParameterVector

from qiskit.circuit.library import EfficientSU2, PauliTwoDesign, RealAmplitudes


class QuantumCircuitLibrary(object):
    """Module for storing various quantum circuits."""

    def __init__(self, n_qubits=3, reps=1):
        """Initialize class."""
        self.n_qubits = n_qubits
        self.reps = reps

    def circuit1(self):
        """Generate tight-binding ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        circ = QuantumCircuit(n_qubits)
        params = ParameterVector("θ", reps * 1 * n_qubits)
        count = 0
        for r in range(reps):
            for i in range(n_qubits):
                circ.ry(params[count], i)
                count += 1
        return circ

    def circuit2(self):
        """Generate tight-binding ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        circ = QuantumCircuit(n_qubits)
        params = ParameterVector("θ", reps * 2 * n_qubits)
        count = 0
        for r in range(reps):

            for i in range(n_qubits):
                circ.ry(params[count], i)
                count += 1
                circ.rz(params[count], i)
                count += 1
        return circ

    def circuit3(self):
        """Generate tight-binding ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        circ = QuantumCircuit(n_qubits)
        params = ParameterVector("θ", reps * 2 * n_qubits)
        count = 0
        for r in range(reps):

            for i in range(n_qubits):
                circ.ry(params[count], i)
                count += 1
                circ.rz(params[count], i)
                count += 1
            for i in range(n_qubits):

                if i != n_qubits - 1:
                    circ.cx(i, n_qubits - 1)

        return circ

    def circuit4(self):
        """Generate RealAmplitudes ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        return RealAmplitudes(num_qubits=n_qubits, reps=reps)

    def circuit5(self):
        """Generate PauliTwoDesign ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        return PauliTwoDesign(num_qubits=n_qubits, reps=reps)

    def circuit6(self):
        """Generate EfficientSU2 ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        return EfficientSU2(num_qubits=n_qubits, reps=reps)

    def circuit7(self):
        """Generate tight-binding ansatz."""
        reps = self.reps
        n_qubits = self.n_qubits
        circ = QuantumCircuit(n_qubits)
        params = ParameterVector("theta", reps * 4 * n_qubits)
        count = 0
        for r in range(reps):

            for i in range(n_qubits):
                circ.ry(params[count], i)
                count += 1
                circ.rz(params[count], i)
                count += 1
            for i in range(n_qubits):

                if i != n_qubits - 1:
                    circ.cx(i, n_qubits - 1)

                for j in range(n_qubits):
                    if i + j < n_qubits and j > 0:
                        circ.cx(i, i + j)
            for i in range(n_qubits):
                circ.ry(params[count], i)
                count += 1
                circ.rz(params[count], i)
                count += 1
        return circ


"""
if __name__ == "__main__":
    from jarvis.db.figshare import (
        get_wann_phonon,
        get_hk_tb,
        get_wann_electron,
    )
    from jarvis.io.qiskit.inputs import HermitianSolver

    w, atoms, ef = get_wann_electron("JVASP-816")
    kps = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    for kk, k in enumerate(kps):
        print()
        hk = get_hk_tb(w=w, k=k)
        HS = HermitianSolver(hk)
        n_qubits = HS.n_qubits()
        Q = QuantumCircuitLibrary(n_qubits=n_qubits)
        circs = [
            Q.circuit1(),
            Q.circuit2(),
            Q.circuit3(),
            Q.circuit4(),
            Q.circuit5(),
            Q.circuit6(),
        ]
        for ii, i in enumerate(circs):
            en, vqe_result, vqe = HS.run_vqe(var_form=i)
            vals, vecs = HS.run_numpy()
            diff = abs(en - vals[0])
            print("VQE,numpy", ii, kk, en, vals[0], diff)
            print(i)
"""



"""Modules related to chemistry of periodic-table elements."""

import os
import json
import numpy as np
import functools
from jarvis.core.utils import digitize_array
from collections import defaultdict
from collections.abc import Iterable
from jarvis.db.jsonutils import loadjson

el_chem_json_file = str(
    os.path.join(os.path.dirname(__file__), "Elements.json")
)
el_chem_json = open(el_chem_json_file, "r")
chem_data = json.load(el_chem_json)
el_chem_json.close()

chem_data_magpie = loadjson(
    os.path.join(os.path.dirname(__file__), "magpie.json")
)

el_chrg_json_file = str(
    os.path.join(os.path.dirname(__file__), "element_charge.json")
)
el_chrg_json = open(el_chrg_json_file, "r")
chrg_data = json.load(el_chrg_json)
el_chrg_json.close()
cgcnn_feature_json = os.path.join(os.path.dirname(__file__), "atom_init.json")


def get_descrp_arr_name(elm="Al"):
    """
    Get chemical descriptors for an element.

    Can be used in JARVIS-ML.

    Args:

           elm: element name

    Returns:
             arr: array value
    """
    arr = []
    try:
        dat = chem_data
        d = dat[elm]
        arr = []
        for k, v in d.items():
            arr.append(k)
    except Exception:
        pass
    return arr


class Specie(object):
    """
    Specie object for chemistry information.

    Used in defining chemistry of a material.

    >>> el = Specie('Al')
    >>> el.Z
    13
    >>> round(el.atomic_mass,2)
    26.98
    >>> el.symbol
    'Al'
    >>> round(el.get_chgdescrp_arr[1],2)
    12.17
    >>> round(el.get_descrp_arr[1],2)
    2792.11
    >>> el = Specie('asdfg')
    >>> el.element_property("asdfg")
    nan

    """

    def __init__(self, symbol="", source="cfid"):
        """Initialize with periodic table element."""
        self.symbol = symbol
        if source == "cfid":
            # Cite reference:
            # https://doi.org/10.1103/PhysRevMaterials.2.083801
            self._data = chem_data
        elif source == "magpie":
            # Cite reference:
            # https://doi.org/10.1038/npjcompumats.2016.28
            self._data = chem_data_magpie
        else:
            raise ValueError("Option not available.", source)

    @property
    def Z(self):
        """Get atomic number."""
        return self.element_property("Z")

    @property
    def atomic_mass(self):
        """Get atomic mass."""
        return self.element_property("atom_mass")

    @property
    def get_chgdescrp_arr(self):
        """
        Get charge descriptors for an element.

        Gives 378 array data.

        Args:

           elm: element name

        Returns:
             arr: array value
        """
        arr = []

        # try:
        arr = chrg_data[self.symbol][0][1]
        # except:
        #    pass
        return arr

    @property
    def get_descrp_arr(self):
        """
        Get chemical descriptors for an element.

        Gives 438 array data.

        Args:

           elm: element name

        Returns:
             arr: array value
        """
        arr = []
        # print ('self._data')
        d = self._data[self.symbol]
        # d = chem_data[self.symbol]
        arr = []
        for k, v in d.items():
            arr.append(v)
        arr = np.array(arr).astype(float)
        return arr

    @property
    def atomic_rad(self):
        """Get atomic radii."""
        return self.element_property("atom_rad")

    @property
    def X(self):
        """Get electronegativity."""
        return self.element_property("X")

    def element_property(self, key=""):
        """
        Get element property from the list of keys.

        These 84 keys are:
        keys = [
                "is_halogen",
                "row",
                "GV",
                "nfunfill",
                "C-9",
                "C-8",
                "C-7",
                "C-6",
                "C-5",
                "C-4",
                "C-3",
                "C-2",
                "C-1",
                "C-0",
                "me1",
                "me3",
                "me2",
                "max_oxid_s",
                "npvalence",
                "mp",
                "first_ion_en",
                "ndunfill",
                "op_eg",
                "jv_enp",
                "nfvalence",
                "polzbl",
                "oq_bg",
                "atom_rad",
                "atom_mass",
                "is_alkali",
                "C-13",
                "C-12",
                "C-11",
                "C-10",
                "C-17",
                "C-16",
                "C-15",
                "C-14",
                "C-19",
                "C-18",
                "voro_coord",
                "is_noble_gas",
                "e1",
                "e3",
                "e2",
                "is_lanthanoid",
                "ndvalence",
                "KV",
                "min_oxid_s",
                "nsunfill",
                "C-26",
                "X",
                "is_actinoid",
                "C-28",
                "C-29",
                "C-27",
                "C-24",
                "C-25",
                "C-22",
                "C-23",
                "C-20",
                "C-21",
                "avg_ion_rad",
                "nsvalence",
                "is_metalloid",
                "elec_aff",
                "coulmn",
                "mol_vol",
                "bp",
                "C-31",
                "C-30",
                "C-33",
                "C-32",
                "C-35",
                "C-34",
                "is_transition_metal",
                "block",
                "therm_cond",
                "Z",
                "is_alkaline",
                "npunfill",
                "oq_enp",
                "mop_eg",
                "hfus",
            ]
        """
        val = np.nan
        try:
            val = self._data[self.symbol][key]
        except Exception:
            pass
        return val


BASIC_FEATURES = [
    "Z",
    "coulmn",
    "row",
    "X",
    "atom_rad",
    "nsvalence",
    "npvalence",
    "ndvalence",
    "nfvalence",
    "first_ion_en",
    "elec_aff",
]


@functools.lru_cache(maxsize=None)
def get_node_attributes(species, atom_features="atomic_number"):
    """Get specific node features for an element."""
    feature_sets = ("atomic_number", "basic", "cfid", "cgcnn")

    if isinstance(atom_features, str):
        if atom_features not in feature_sets:
            raise NotImplementedError(
                f"atom features must be one of {feature_sets}"
            )
    elif isinstance(atom_features, Iterable):
        # allow custom list of features
        for prop in atom_features:
            if prop not in keys:
                raise NotImplementedError(
                    f"{prop} not supported in custom atom feature list"
                )
        return [
            Specie(species).element_property(prop) for prop in atom_features
        ]

    if atom_features == "cfid":
        return Specie(species).get_descrp_arr
    elif atom_features == "atomic_number":
        return [Specie(species).element_property("Z")]
    elif atom_features == "basic":
        return [
            Specie(species).element_property(prop) for prop in BASIC_FEATURES
        ]
    elif atom_features == "cgcnn":
        # load from json, key by atomic number
        key = str(Specie(species).element_property("Z"))
        with open(cgcnn_feature_json, "r") as f:
            # For alternative features use
            # get_digitized_feats_hot_encoded()
            i = json.load(f)
        try:
            return i[key]
        except KeyError:
            print(f"warning: could not load CGCNN features for {key}")
            print("Setting it to max atomic number available here, 103")
            # TODO Check for the error in oqmd_3d_no_cfid dataset
            # return i['Lr']
            return i["100"]


keys = [
    "is_halogen",
    "row",
    "GV",
    "nfunfill",
    "C-9",
    "C-8",
    "C-7",
    "C-6",
    "C-5",
    "C-4",
    "C-3",
    "C-2",
    "C-1",
    "C-0",
    "me1",
    "me3",
    "me2",
    "max_oxid_s",
    "npvalence",
    "mp",
    "first_ion_en",
    "ndunfill",
    "op_eg",
    "jv_enp",
    "nfvalence",
    "polzbl",
    "oq_bg",
    "atom_rad",
    "atom_mass",
    "is_alkali",
    "C-13",
    "C-12",
    "C-11",
    "C-10",
    "C-17",
    "C-16",
    "C-15",
    "C-14",
    "C-19",
    "C-18",
    "voro_coord",
    "is_noble_gas",
    "e1",
    "e3",
    "e2",
    "is_lanthanoid",
    "ndvalence",
    "KV",
    "min_oxid_s",
    "nsunfill",
    "C-26",
    "X",
    "is_actinoid",
    "C-28",
    "C-29",
    "C-27",
    "C-24",
    "C-25",
    "C-22",
    "C-23",
    "C-20",
    "C-21",
    "avg_ion_rad",
    "nsvalence",
    "is_metalloid",
    "elec_aff",
    "coulmn",
    "mol_vol",
    "bp",
    "C-31",
    "C-30",
    "C-33",
    "C-32",
    "C-35",
    "C-34",
    "is_transition_metal",
    "block",
    "therm_cond",
    "Z",
    "is_alkaline",
    "npunfill",
    "oq_enp",
    "mop_eg",
    "hfus",
]


def get_specie_data():
    """Get the json and key data from Specie."""
    return keys, chem_data, chrg_data


def get_digitized_feats_hot_encoded(
    feature_names=keys, filename="feats_encoded.json"
):
    """Get OneHotEncoded features with digitized features."""
    from sklearn.preprocessing import OneHotEncoder
    import pandas as pd

    encoder = OneHotEncoder(categories="auto", sparse_output=False)
    # encoder = OneHotEncoder(categories="auto", sparse=False)
    dat = defaultdict()
    for i, j in chem_data.items():
        tmp = defaultdict()
        for r, s in j.items():
            if r in feature_names:
                tmp[r] = s
        dat[Specie(i).Z] = tmp  # j.values()
    df = pd.DataFrame(dat)
    df = df.T.replace(-9999.0, 0).replace(-0.0, 0).astype("float")

    for i in df.columns:
        df[i] = digitize_array(df[i])
    df = df.T

    vals = []
    for i in range(len(df.values)):
        output = encoder.fit_transform(
            np.array(df.values[i], dtype="float").reshape(-1, 1)
        )  # .toarray()
        vals.extend(output.T)
    vals = np.array(vals, dtype="float").T
    cols = df.columns.tolist()
    new_dat = {}
    for i, j in zip(cols, vals):
        new_dat[int(i)] = list([int(m) for m in j])
    if filename is not None:
        from jarvis.db.jsonutils import dumpjson

        dumpjson(data=new_dat, filename=filename)
    return new_dat


def get_feats_hot_encoded(feature_names=keys, filename="feats_encoded.json"):
    """Get OneHotEncoded features."""
    # Deprecated
    # Kept for reference only
    from sklearn.preprocessing import OneHotEncoder
    import pandas as pd

    encoder = OneHotEncoder(categories="auto", sparse_output=False)
    # encoder = OneHotEncoder(categories="auto", sparse=False)
    dat = {}
    for i, j in chem_data.items():
        tmp = []
        for r, s in j.items():
            if r in feature_names:
                tmp.append(s)
        dat[Specie(i).Z] = tmp  # j.values()
    df = pd.DataFrame(dat)

    vals = []
    for i in range(len(df.values)):
        output = encoder.fit_transform(
            np.array(df.values[i], dtype="float").reshape(-1, 1)
        )  # .toarray()
        vals.extend(output.T)
    vals = np.array(vals, dtype="float").T
    cols = df.columns.tolist()
    new_dat = {}
    for i, j in zip(cols, vals):
        new_dat[i] = list(j)
    if filename is not None:
        from jarvis.db.jsonutils import dumpjson

        dumpjson(data=new_dat, filename=filename)
    return new_dat


x, y, z = get_specie_data()
info_z = {}
for i, j in y.items():
    info_z[j["Z"]] = i


def atomic_numbers_to_symbols(numbers=[1, 2, 3, 4]):
    """Convert atomic number array to atomic symbols."""
    symbs = []
    for i in numbers:
        symbs.append(info_z[i])
    return symbs


# get_digitized_feats_hot_encoded()
"""
if __name__ == "__main__":
    el = Specie("Al")
    #print(el.get_chgdescrp_arr)
    print(len(el.get_descrp_arr))
    #print (get_descrp_arr_name())
"""



"""Module to prepare input files."""


def get_Zeff():
    """Get valence electrons."""
    return dict(
        Ag=19,
        Al=3,
        Ar=8,
        As=5,
        Au=19,
        B=3,
        Be=2,
        Bi=5,
        Br=7,
        C=4,
        Ca=10,
        Cl=7,
        Co=17,
        Cr=14,
        Cu=19,
        F=7,
        Fe=16,
        Ga=3,
        Ge=4,
        H=1,
        He=2,
        I=7,
        Ir=17,
        K=9,
        Kr=8,
        Li=1,
        Mg=2,
        Mn=15,
        Mo=14,
        N=5,
        Na=1,
        Ne=8,
        Ni=18,
        O=6,
        P=5,
        Pd=18,
        S=6,
        Sc=11,
        Se=6,
        Si=4,
        Tb=19,
        Te=6,
        Ti=12,
        V=13,
        W=14,
        Zn=20,
    )


def get_pseudo_dft_dict():
    """Get PSPs for DFT."""
    return {
        "Ag": "Ag.ccECP.AREP.upf",  # Ag.ccECP.SOREP.upf
        "Al": "Al.ccECP.upf",
        "Ar": "Ar.ccECP.upf",
        "As": "As.ccECP.upf",
        "Au": "Au.ccECP.AREP.upf",  # Au.ccECP.SOREP.upf
        "B": "B.ccECP.upf",
        "Be": "Be.ccECP.upf",
        "Bi": "Bi.ccECP.AREP.upf",  # Bi.ccECP.SOREP.upf
        "Br": "Br.ccECP.upf",
        "C": "C.ccECP.upf",
        "Ca": "Ca.ccECP.upf",
        "Cl": "Cl.ccECP.upf",
        "Co": "Co.ccECP-soft.upf",  # Co.opt.upf
        "Cr": "Cr.ccECP.upf",  # Cr.opt.upf
        "Cu": "Cu.ccECP-soft.upf",  # Cu.opt.upf
        "F": "F.ccECP.upf",
        "Fe": "Fe.ccECP-soft.upf",  # Fe.opt.upf
        "Ga": "Ga.ccECP.upf",
        "Ge": "Ge.ccECP.upf",
        "H": "H.ccECP.upf",
        "He": "He.ccECP.upf",
        "I": "I.ccECP.AREP.upf",  # I.ccECP.SOREP.upf
        "Ir": "Ir.ccECP.AREP.upf",  # Ir.ccECP.SOREP.upf
        "K": "K.ccECP.upf",
        "Kr": "Kr.ccECP.upf",
        "Li": "Li.ccECP.upf",
        "Mg": "Mg.ccECP.upf",
        "Mn": "Mn.ccECP.upf",  # Mn.opt.upf
        "Mo": "Mo.ccECP.AREP.upf",  # Mo.ccECP.SOREP.upf
        "N": "N.ccECP.upf",
        "Na": "Na.ccECP.upf",
        "Ne": "Ne.ccECP.upf",
        "Ni": "Ni.ccECP-soft.upf",  # Ni.opt.upf
        "O": "O.ccECP.upf",
        "P": "P.ccECP.upf",
        "Pd": "Pd.ccECP.AREP.upf",  # Pd.ccECP.SOREP.upf
        "S": "S.ccECP.upf",
        "Sc": "Sc.ccECP-soft.upf",  # Sc.opt.upf
        "Se": "Se.ccECP.upf",
        "Si": "Si.ccECP.upf",
        "Tb": "Tb.ccECP.AREP.upf",  # Tb.ccECP.SOREP.upf
        "Te": "Te.ccECP.AREP.upf",  # Te.ccECP.SOREP.upf
        "Ti": "Ti.ccECP-soft.upf",  # Ti.opt.upf
        "V": "V.ccECP-soft.upf",  # V.opt.upf
        "W": "W.ccECP.AREP.upf",  # W.ccECP.SOREP.upf
        "Zn": "Zn.ccECP-soft.upf",  # Zn.opt.upf
    }


def get_pseudo_qmc_dict():
    """Get PSPs for QMC."""
    return {
        "Ag": "Ag.ccECP.AREP.xml",  # Ag.ccECP.SOREP.xml
        "Al": "Al.ccECP.xml",
        "Ar": "Ar.ccECP.xml",
        "As": "As.ccECP.upf",
        "Au": "Au.ccECP.AREP.xml",  # Au.ccECP.SOREP.xml
        "B": "B.ccECP.xml",
        "Be": "Be.ccECP.xml",
        "Bi": "Bi.ccECP.AREP.xml",  # Bi.ccECP.SOREP.xml
        "Br": "Br.ccECP.xml",
        "C": "C.ccECP.xml",
        "Ca": "Ca.ccECP.xml",
        "Cl": "Cl.ccECP.xml",
        "Co": "Co.ccECP-soft.xml",  # Co.opt.xml
        "Cr": "Cr.ccECP.xml",  # Cr.opt.xml
        "Cu": "Cu.ccECP-soft.xml",  # Cu.opt.xml
        "F": "F.ccECP.xml",
        "Fe": "Fe.ccECP-soft.xml",  # Fe.opt.xml
        "Ga": "Ga.ccECP.xml",
        "Ge": "Ge.ccECP.xml",
        "H": "H.ccECP.xml",
        "He": "He.ccECP.xml",
        "I": "I.ccECP.AREP.xml",  # I.ccECP.SOREP.xml
        "Ir": "Ir.ccECP.AREP.xml",  # Ir.ccECP.SOREP.xml
        "K": "K.ccECP.xml",
        "Kr": "Kr.ccECP.xml",
        "Li": "Li.ccECP.xml",
        "Mg": "Mg.ccECP.xml",
        "Mn": "Mn.ccECP.xml",  # Mn.opt.xml
        "Mo": "Mo.ccECP.AREP.xml",  # Mo.ccECP.SOREP.xml
        "N": "N.ccECP.xml",
        "Na": "Na.ccECP.xml",
        "Ne": "Ne.ccECP.xml",
        "Ni": "Ni.ccECP-soft.xml",  # Ni.opt.xml
        "O": "O.ccECP.xml",
        "P": "P.ccECP.upf",
        "Pd": "Pd.ccECP.AREP.xml",  # Pd.ccECP.SOREP.xml
        "S": "S.ccECP.xml",
        "Sc": "Sc.ccECP-soft.xml",  # Sc.opt.xml
        "Se": "Se.ccECP.xml",
        "Si": "Si.ccECP.xml",
        "Tb": "Tb.ccECP.AREP.xml",  # Tb.ccECP.SOREP.xml
        "Te": "Te.ccECP.AREP.xml",  # Te.ccECP.SOREP.xml
        "Ti": "Ti.ccECP-soft.xml",  # Ti.opt.xml
        "V": "V.ccECP-soft.xml",  # V.opt.xml
        "W": "W.ccECP.AREP.xml",  # W.ccECP.SOREP.xml
        "Zn": "Zn.ccECP-soft.xml",  # Zn.opt.xml
    }
